@@ -35,6 +35,8 @@ static void libbalsa_address_class_init(LibBalsaAddressClass * klass);
 static void libbalsa_address_init(LibBalsaAddress * ab);
 static void libbalsa_address_finalize(GObject * object);
 
+static gchar ** vcard_strsplit(const gchar * string);
+
 GType libbalsa_address_get_type(void)
 {
     static GType address_type = 0;
@@ -111,6 +113,401 @@ libbalsa_address_new(void)
     return g_object_new(LIBBALSA_TYPE_ADDRESS, NULL);
 }
 
+/** Extract full name in order from <string> that has GnomeCard format
+   and returns the pointer to the allocated memory chunk.
+
+   VCARD code attempts to obey published documentation:
+
+   [1] VCARD 1.2 specs: http://www.imc.org/pdi/vcard-21.txt
+   [2] VCARD 3.0 specs, RFC 2426 (http://www.ietf.org/rfc/rfc2426.txt)
+*/
+gchar*
+libbalsa_address_extract_name(const gchar * string, gchar ** last_name,
+                              gchar ** first_name)
+{
+    enum GCardFieldOrder { LAST = 0, FIRST, MIDDLE, PREFIX, SUFFIX };
+    gint cpt, j;
+    gchar **fld, **name_arr;
+    gchar *res = NULL;
+
+    fld = vcard_strsplit(string);
+
+    cpt = 0;
+    while (fld[cpt] != NULL)
+	cpt++;
+
+    if (cpt == 0)		/* insane empty name */
+	return NULL;
+
+    if (fld[LAST] && *fld[LAST])
+        *last_name = g_strdup(fld[LAST]);
+
+    if (fld[FIRST] && *fld[FIRST])
+        *first_name = fld[MIDDLE] && *fld[MIDDLE] ?
+            g_strconcat(fld[FIRST], " ", fld[MIDDLE], NULL) :
+            g_strdup(fld[FIRST]);
+
+    name_arr = g_malloc((cpt + 1) * sizeof(gchar *));
+
+    j = 0;
+    if (cpt > PREFIX && *fld[PREFIX] != '\0')
+	name_arr[j++] = g_strdup(fld[PREFIX]);
+
+    if (cpt > FIRST && *fld[FIRST] != '\0')
+	name_arr[j++] = g_strdup(fld[FIRST]);
+
+    if (cpt > MIDDLE && *fld[MIDDLE] != '\0')
+	name_arr[j++] = g_strdup(fld[MIDDLE]);
+
+    if (cpt > LAST && *fld[LAST] != '\0')
+	name_arr[j++] = g_strdup(fld[LAST]);
+
+    if (cpt > SUFFIX && *fld[SUFFIX] != '\0')
+	name_arr[j++] = g_strdup(fld[SUFFIX]);
+
+    name_arr[j] = NULL;
+
+    g_strfreev(fld);
+
+    /* collect the data to one string */
+    res = g_strjoinv(" ", name_arr);
+    g_strfreev(name_arr);
+
+    return res;
+}
+
+/* Duplicates some code from address-book-vcard.c:
+ */
+static gchar*
+validate_vcard_string(gchar * vcstr, const gchar * charset)
+{
+    gchar * utf8res;
+    gsize b_written;
+
+    /* check if it's a utf8 clean string and return it in this case */
+    if (!vcstr || g_utf8_validate(vcstr, -1, NULL))
+	return vcstr;
+
+    /* convert from the passed charset or as fallback from the locale setting */
+    if (charset && g_ascii_strcasecmp(charset, "utf-8")) {
+	utf8res = g_convert(vcstr, -1, "utf-8", charset, NULL, &b_written, NULL);
+    } else
+	utf8res = g_locale_to_utf8(vcstr, -1, NULL, &b_written, NULL);
+    if (!utf8res)
+	return vcstr;
+
+    g_free(vcstr);
+    return utf8res;
+}
+
+
+static inline gchar *
+vcard_qp_decode(gchar * str)
+{
+    gint len = strlen(str);
+    gchar * newstr = g_malloc0(len + 1);
+    int state = 0;
+#if HAVE_GMIME_2_2_5
+    guint32 save;
+#else  /* HAVE_GMIME_2_2_5 */
+    int save;
+#endif /* HAVE_GMIME_2_2_5 */
+
+    /* qp decode the input string */
+    g_mime_utils_quoted_decode_step((unsigned char *) str, len,
+				    (unsigned char *) newstr, &state, &save);
+
+    /* free input and return new string */
+    g_free(str);
+    return newstr;
+}
+
+
+static inline gchar *
+vcard_b64_decode(gchar * str)
+{
+    gint len = strlen(str);
+    gchar * newstr = g_malloc0(len + 1);
+    int state = 0;
+    guint32 save;
+
+    /* base64 decode the input string */
+    g_mime_utils_base64_decode_step((unsigned char *) str, len,
+				    (unsigned char *) newstr, &state, &save);
+
+    /* free input and return new string */
+    g_free(str);
+    return newstr;
+}
+
+
+static inline gchar *
+vcard_charset_to_utf8(gchar * str, const gchar * charset)
+{
+    gsize bytes_written;
+    gchar * convstr;
+
+    /* convert only if the source is known and not utf-8 */
+    if (!charset || !g_ascii_strcasecmp(charset, "utf-8"))
+	return str;
+
+    convstr = g_convert(str, -1, "utf-8", charset, NULL, &bytes_written, NULL);
+    g_free(str);
+    return convstr ? convstr : strdup("");
+}
+
+
+/* mainly copied from g_strsplit, but (a) with the fixed delimiter ';'
+ * (b) ignoring '\;' sequences (c) always returning as many elements as
+ * possible and (d) unescape '\;' sequences in the resulting array */
+static gchar **
+vcard_strsplit(const gchar * string)
+{
+    GSList *string_list = NULL, *slist;
+    gchar **str_array, *s;
+    guint n = 0;
+    const gchar *remainder;
+    gint max_tokens = G_MAXINT;
+
+    g_return_val_if_fail(string != NULL, NULL);
+
+    max_tokens = G_MAXINT;
+    remainder = string;
+    s = strchr(remainder, ';');
+    while (s && s > remainder && s[-1] == '\\')
+	s = strchr(s + 1, ';');
+
+    while (s) {
+	gsize len;
+
+	len = s - remainder;
+	string_list = g_slist_prepend(string_list,
+				      g_strndup(remainder, len));
+	n++;
+	remainder = s + 1;
+	s = strchr(remainder, ';');
+	while (s && s > remainder && s[-1] == '\\')
+	    s = strchr(s + 1, ';');
+    }
+
+    if (*string) {
+	n++;
+	string_list = g_slist_prepend(string_list, g_strdup(remainder));
+    }
+
+    str_array = g_new(gchar*, n + 1);
+
+    str_array[n--] = NULL;
+    for (slist = string_list; slist; slist = slist->next) {
+	gchar * str = (gchar *) slist->data;
+	gchar * p;
+
+	while ((p = strstr(str, "\\;"))) {
+	    gchar * newstr = g_malloc(strlen(str));
+
+	    strncpy(newstr, str, p - str);
+	    strcpy(newstr + (p - str), p + 1);
+	    g_free(str);
+	    str = newstr;
+	}
+	str_array[n--] = str;
+    }
+
+    g_slist_free(string_list);
+
+    return str_array;
+}
+
+
+LibBalsaAddress*
+libbalsa_address_new_from_vcard(const gchar *str, const gchar *charset)
+{
+    LibBalsaAddress *address = g_object_new(LIBBALSA_TYPE_ADDRESS, NULL);
+    gchar *name = NULL, *nick_name = NULL, *org = NULL;
+    gchar *full_name = NULL, *last_name = NULL, *first_name = NULL;
+    gint in_vcard = FALSE;
+    GList *address_list = NULL;
+    const gchar *string, *next_line;
+    gchar * vcard;
+
+    g_return_val_if_fail(str, address);
+
+    /* rfc 2425 unfold the string */
+    vcard = g_strdup(str);
+    while ((string = strstr(vcard, "\r\n ")) ||
+	   (string = strstr(vcard, "\r\n\t"))) {
+	gchar * newstr = g_malloc0(strlen(vcard) - 2);
+
+	strncpy(newstr, vcard, string - vcard);
+	strcpy(newstr + (string - vcard), string + 3);
+	g_free(vcard);
+	vcard = newstr;
+    }
+    while ((string = strstr(vcard, "\n ")) ||
+	   (string = strstr(vcard, "\n\t"))) {
+	gchar * newstr = g_malloc(strlen(vcard) - 1);
+
+	strncpy(newstr, vcard, string - vcard);
+	strcpy(newstr + (string - vcard), string + 2);
+	g_free(vcard);
+	vcard = newstr;
+    }
+
+    /* may contain \r's when decoded from base64... */
+    while ((string = strstr(vcard, "\r\n ")) ||
+	   (string = strstr(vcard, "\r\n\t"))) {
+	gchar * newstr = g_malloc(strlen(vcard) - 2);
+
+	strncpy(newstr, vcard, string - vcard);
+	strcpy(newstr + (string - vcard), string + 3);
+	g_free(vcard);
+	vcard = newstr;
+    }
+
+    /* process */
+    for(string = vcard; *string; string = next_line) {
+        next_line = strchr(string, '\n');
+        if (next_line)
+            ++next_line;
+        else
+            next_line = string + strlen(string);
+	/*
+	 * Check if it is a card.
+	 */
+	if (g_ascii_strncasecmp(string, "BEGIN:VCARD", 11) == 0) {
+	    in_vcard = TRUE;
+	} else if (g_ascii_strncasecmp(string, "END:VCARD", 9) == 0) {
+            /*
+             * We are done loading a card.
+             */
+	    if (address_list) {
+                if (address) {
+                    if (full_name) {
+                        address->full_name = full_name;
+                        g_free(name);
+                    } else if (name)
+                        address->full_name = name;
+                    else if (nick_name)
+                        address->full_name = g_strdup(nick_name);
+                    else
+                        address->full_name = g_strdup(_("No-Name"));
+
+                    address->last_name = last_name;
+                    address->first_name = first_name;
+                    address->nick_name = nick_name;
+                    address->organization = org;
+                    address->address_list = g_list_reverse(address_list);
+
+                    break;
+                }
+                g_list_foreach(address_list, (GFunc) g_free, NULL);
+                g_list_free(address_list);
+	    }
+            /* Record without e-mail address, or we're not creating
+             * addresses: free memory. */
+            g_free(full_name);
+            g_free(name);
+            g_free(last_name);
+            g_free(first_name);
+            g_free(nick_name);
+            g_free(org);
+            
+            /* We return just one address... */
+            return address;
+	} else if (in_vcard) {
+            gchar *line = g_strndup(string, next_line-string);
+
+            g_strchomp(line);
+
+	    /* Encoding of national characters:
+	     * - vcard 2.1 allows charset=xxx and encoding=(base64|quoted-printable|8bit)
+	     * - Thunderbird claims to use vcard 2.1, but uses only "quoted-printable",
+	     *   and the charset from part MIME header
+	     * - vcard 3.0 (rfc 2426) allows "encoding=b", the charset must be taken
+	     *   from the content-type MIME header */
+	    if (strchr(line, ':')) {
+		gchar ** parts = g_strsplit(line, ":", 2);
+		gchar ** tokens;
+		gint n;
+
+		/* split control stuff into tokens */
+		tokens = g_strsplit(parts[0], ";", -1);
+
+		/* find encoding= */
+		for (n = 0; tokens[n]; n++)
+		    if (!g_ascii_strncasecmp(tokens[n], "encoding=", 9)) {
+			if (!g_ascii_strcasecmp(tokens[n] + 9, "base64")) {
+			    /* vcard 2.1: use the charset parameter (below) */
+			    parts[1] = vcard_b64_decode(parts[1]);
+			} else if (!g_ascii_strcasecmp(tokens[n] + 9, "b")) {
+			    /* rfc 2426: charset from MIME part */
+			    parts[1] = vcard_b64_decode(parts[1]);
+			    parts[1] = vcard_charset_to_utf8(parts[1], charset);
+			} else if (!g_ascii_strcasecmp(tokens[n] + 9, "quoted-printable")) {
+			    /* vcard 2.1: use the charset parameter (below) */
+			    parts[1] = vcard_qp_decode(parts[1]);
+			}
+		    }
+
+		/* find quoted-printable */
+		for (n = 0; tokens[n]; n++)
+		    if (!g_ascii_strcasecmp(tokens[n], "quoted-printable")) {
+			/* Thunderbird: broken vcard 2.1, charset from MIME part */
+			parts[1] = vcard_qp_decode(parts[1]);
+			parts[1] = vcard_charset_to_utf8(parts[1], charset);
+		    }
+
+		/* find charset= (vcard 2.1 only) */
+		for (n = 0; tokens[n]; n++)
+		    if (!g_ascii_strncasecmp(tokens[n], "charset=", 8))
+			parts[1] = vcard_charset_to_utf8(parts[1], tokens[n] + 8);
+
+		/* construct the result */
+		g_free(line);
+		line = g_strdup_printf("%s:%s", tokens[0], parts[1]);
+
+		/* clean up */
+		g_strfreev(tokens);
+		g_strfreev(parts);
+	    }
+
+            if (g_ascii_strncasecmp(line, "FN:", 3) == 0) {
+
+                full_name = g_strdup(line + 3);
+                full_name = validate_vcard_string(full_name, charset);
+
+            } else if (g_ascii_strncasecmp(line, "N:", 2) == 0) {
+
+                name = libbalsa_address_extract_name(line + 2,
+                                                     &last_name, &first_name);
+                name = validate_vcard_string(name, charset);
+                last_name = validate_vcard_string(last_name, charset);
+                first_name = validate_vcard_string(first_name, charset);
+
+            } else if (g_ascii_strncasecmp(line, "NICKNAME:", 9) == 0) {
+
+                nick_name = g_strdup(line + 9);
+                nick_name = validate_vcard_string(nick_name, charset);
+
+            } else if (g_ascii_strncasecmp(line, "ORG:", 4) == 0) {
+		gchar ** org_strs = vcard_strsplit(line + 4);
+
+                org = g_strjoinv(", ", org_strs);
+		g_strfreev(org_strs);
+                org = validate_vcard_string(org, charset);
+
+            } else if (g_ascii_strncasecmp(line, "EMAIL:", 6) == 0) {
+
+		address_list =
+		    g_list_prepend(address_list, g_strdup(line + 6));
+            }
+            g_free(line);
+        }
+    }
+
+    return address;
+}
+
 void
 libbalsa_address_set_copy(LibBalsaAddress * dest, LibBalsaAddress * src)
 {
@@ -255,7 +652,7 @@ libbalsa_address_get_mailbox_from_list(const InternetAddressList *
     in entries with values from address
 */
 void
-libbalsa_address_set_edit_entries(LibBalsaAddress * address,
+libbalsa_address_set_edit_entries(const LibBalsaAddress * address,
                                   GtkWidget **entries)
 {
     gchar *new_name = NULL;
@@ -417,8 +814,12 @@ add_row(GtkWidget*button, gpointer data)
     GtkListStore *store = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
     GtkTreeIter iter;
     GtkTreePath *path;
-    gtk_list_store_insert_with_values(store, &iter, 99999,
-                                      0, "", -1);
+#if GTK_CHECK_VERSION(2, 6, 0)
+    gtk_list_store_insert_with_values(store, &iter, 99999, 0, "", -1);
+#else                           /* GTK_CHECK_VERSION(2, 6, 0) */
+    gtk_list_store_append(store, &iter);
+    gtk_list_store_set(store, &iter, 0, "", -1);
+#endif                          /* GTK_CHECK_VERSION(2, 6, 0) */
     gtk_widget_grab_focus(GTK_WIDGET(tv));
     path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
     gtk_tree_view_set_cursor(tv, path, NULL, TRUE);
@@ -449,11 +850,18 @@ addrlist_drag_received_cb(GtkWidget * widget, GdkDragContext * context,
             addr = *(LibBalsaAddress**)selection_data->data;
             if(addr && addr->address_list) {
                 g_print ("string: %s\n", (gchar*)addr->address_list->data);
+#if GTK_CHECK_VERSION(2, 6, 0)
                 gtk_list_store_insert_with_values(GTK_LIST_STORE(model),
                                                   &iter, 99999,
                                                   0,
                                                   addr->address_list->data,
                                                   -1);
+#else                           /* GTK_CHECK_VERSION(2, 6, 0) */
+                gtk_list_store_append(GTK_LIST_STORE(model), &iter);
+                gtk_list_store_set(GTK_LIST_STORE(model), &iter,
+                                   0, addr->address_list->data,
+                                   -1);
+#endif                          /* GTK_CHECK_VERSION(2, 6, 0) */
                 dnd_success = TRUE;
             }
             break;
@@ -504,7 +912,8 @@ addrlist_drag_drop_cb(GtkWidget *widget, GdkDragContext *context,
 
 
 GtkWidget*
-libbalsa_address_get_edit_widget(LibBalsaAddress *address, GtkWidget **entries,
+libbalsa_address_get_edit_widget(const LibBalsaAddress *address,
+                                 GtkWidget **entries,
                                  GCallback changed_cb, gpointer changed_data)
 {
     const static gchar *labels[NUM_FIELDS] = {
@@ -603,6 +1012,9 @@ libbalsa_address_new_from_edit_entries(GtkWidget ** entries)
     gboolean valid;
     GtkTreeIter iter;
 
+    /* make sure gtk_tree_model looses focus, otherwise the list does
+     * not get updated (gtk2-2.8.20) */
+    gtk_widget_grab_focus(entries[FULL_NAME]);
     /* FIXME: This problem should be solved in the VCard
        implementation in libbalsa: semicolons mess up how GnomeCard
        processes the fields, so disallow them and replace them
