@@ -453,7 +453,7 @@ imap_mbox_append(ImapMboxHandle *handle, const char *mbox,
   gchar *mbx7 = imap_utf8_to_mailbox(mbox);
   gchar *cmd;
   char *litstr = use_literal ? "+" : "";
-  char buf[4096];
+  char buf[8192];
   size_t s, delta;
   int c;
 
@@ -702,10 +702,16 @@ imap_assure_needed_flags(ImapMboxHandle *h, ImapMsgFlag needed_flags)
   int ics;
   gchar *cmd = NULL, *seqno;
   ImapResponse rc = IMR_OK;
+  gchar *cmd_format;
 
   if (h->state == IMHS_DISCONNECTED)
     return IMR_SEVERED;
 
+  if(imap_mbox_handle_can_do(h, IMCAP_ESEARCH))
+    cmd_format = "SEARCH RETURN (ALL) %s %s";
+  else
+    cmd_format = "SEARCH %s %s";
+
   cb  = h->search_cb;  h->search_cb  = (ImapSearchCb)set_flag_cache_cb;
   arg = h->search_arg; 
 
@@ -735,7 +741,7 @@ imap_assure_needed_flags(ImapMboxHandle *h, ImapMsgFlag needed_flags)
       default: g_free(seqno); continue;
       }
       if(!cmd) imap_handle_idle_disable(h);
-      cmd = g_strdup_printf("SEARCH %s %s", seqno, flg);
+      cmd = g_strdup_printf(cmd_format, seqno, flg);
       g_free(seqno);
       flag[issued_cmd] = fnd.flag;
       ics = imap_cmd_start(h, cmd, &cmdno[issued_cmd++]);
@@ -956,6 +962,35 @@ imap_mbox_handle_fetch_rfc822_uid(ImapMboxHandle* handle, unsigned uid,
   return rc;
 }
 
+/** A structure needed to add a faked header to data fetched via
+    binary extension. */
+struct ImapBinaryData {
+  ImapBody *body;
+  ImapFetchBodyCb body_cb;
+  void *body_arg;
+  gboolean first_run;
+};
+
+static void
+imap_binary_handler(unsigned seqno, const char *buf,
+                    int buflen, void* arg)
+{
+  struct ImapBinaryData *ibd = (struct ImapBinaryData*)arg;
+  if(ibd->first_run) {
+    static const char binary_header[] =
+      "Content-Transfer-Encoding: binary\r\n\r\n";
+    char *content_type = imap_body_get_content_type(ibd->body);
+    char *str = g_strdup_printf("Content-Type: %s\r\n", content_type);
+    g_free(content_type);
+    ibd->body_cb(seqno, str, strlen(str), ibd->body_arg);
+    g_free(str);
+    ibd->body_cb(seqno, binary_header, sizeof(binary_header)-1,
+                 ibd->body_arg);
+    ibd->first_run = FALSE;
+  }
+  ibd->body_cb(seqno, buf, buflen, ibd->body_arg);
+}
+
 ImapResponse
 imap_mbox_handle_fetch_body(ImapMboxHandle* handle, 
                             unsigned seqno, const char *section,
@@ -963,15 +998,34 @@ imap_mbox_handle_fetch_body(ImapMboxHandle* handle,
                             ImapFetchBodyCb body_cb, void *arg)
 {
   char cmd[160];
-  ImapMessage *msg;
   ImapFetchBodyCb fcb = handle->body_cb;
   void          *farg = handle->body_arg;
   ImapResponse rc;
 
   IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
+
+  /* Use BINARY extension if possible */
+  if(handle->enable_binary && options == IMFB_MIME &&
+     imap_mbox_handle_can_do(handle, IMCAP_BINARY)) {
+    struct ImapBinaryData ibd;
+    ImapMessage * imsg = imap_mbox_handle_get_msg(handle, seqno);
+    ibd.body = imap_message_get_body_from_section(imsg, section);
+    ibd.body_cb = body_cb;
+    ibd.body_arg = arg;
+    handle->body_cb = imap_binary_handler;
+    handle->body_arg = &ibd;
+    snprintf(cmd, sizeof(cmd), "FETCH %u BINARY[%s]",
+             seqno, section);
+    rc = imap_cmd_exec(handle, cmd);
+    if(rc != IMR_NO) { /* unknown-cte */
+      handle->body_cb  = fcb;
+      handle->body_arg = farg;
+      return rc;
+    }
+  }
   handle->body_cb  = body_cb;
   handle->body_arg = arg;
-  msg = imap_mbox_handle_get_msg(handle, seqno);
+  /* Pure IMAP without extensions */
   if(options == IMFB_NONE)
     snprintf(cmd, sizeof(cmd), "FETCH %u BODY[%s]",
              seqno, section);
@@ -982,7 +1036,7 @@ imap_mbox_handle_fetch_body(ImapMboxHandle* handle,
       unsigned sz;
       char *last_dot = strrchr(section, '.');
       strncpy(prefix, section, sizeof(prefix));
-
+      
       if(last_dot) {
         sz = last_dot-section+1;
         if(sz>sizeof(prefix)-1) sz = sizeof(prefix)-1;
@@ -991,7 +1045,6 @@ imap_mbox_handle_fetch_body(ImapMboxHandle* handle,
       prefix[sizeof(prefix)-1] = '\0';
     } else
       snprintf(prefix, sizeof(prefix), "%s.MIME", section);
-
     snprintf(cmd, sizeof(cmd), "FETCH %u (BODY[%s] BODY[%s])",
              seqno, prefix, section);
   }
@@ -1216,16 +1269,19 @@ sort_code_to_string(ImapSortKey key)
   return keystr;
 }
 
-ImapResponse
-imap_mbox_sort_msgno(ImapMboxHandle *handle, ImapSortKey key,
-                     int ascending, unsigned int *msgno, unsigned cnt)
+/** executes server side sort. The @param msgno array contains @param
+    cnt message numbers. It is subseqently replaced with a new,
+    altered order */
+static ImapResponse
+imap_mbox_sort_msgno_srv(ImapMboxHandle *handle, ImapSortKey key,
+                         int ascending, unsigned int *msgno, unsigned cnt)
 {
   ImapResponse rc;
   const char *keystr;
   char *seq, *cmd, *cmd1;
   unsigned i;
 
-  IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
+  /* IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD); */
   if(!imap_mbox_handle_can_do(handle, IMCAP_SORT)) 
     return IMR_NO;
 
@@ -1265,12 +1321,173 @@ imap_mbox_sort_msgno(ImapMboxHandle *handle, ImapSortKey key,
   return rc;
 }
 
+static int
+comp_unsigned(const void *a, const void *b)
+{
+  return ((const unsigned*)a) - ((const unsigned*)b);
+}
+
+static int
+comp_imap_address(const ImapAddress *a, const ImapAddress *b)
+{
+  if(a && a->name) {
+    if(b && b->name)
+      return g_ascii_strcasecmp(a->name, b->name);
+    else return 1;
+  } else
+    return -1;
+}
+
+struct SortItem {
+  ImapMessage *msg;
+  unsigned no;
+};
+static int
+comp_arrival(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return x->internal_date - y->internal_date;
+}
+static int
+comp_cc(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return comp_imap_address(x->envelope->cc,  y->envelope->cc);
+}
+static int
+comp_date(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return x->envelope->date - y->envelope->date;
+}
+static int
+comp_from(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return comp_imap_address(x->envelope->from,  y->envelope->from);
+}
+static int
+comp_size(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return x->rfc822size - y->rfc822size;
+}
+static int
+comp_subject(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  if(!x->envelope->subject) return -1;
+  if(!y->envelope->subject) return  1;
+  return g_ascii_strcasecmp(x->envelope->subject, y->envelope->subject);
+}
+static int
+comp_to(const void *a, const void *b)
+{
+  const ImapMessage *x = ((const struct SortItem*)a)->msg;
+  const ImapMessage *y = ((const struct SortItem*)b)->msg;
+  return comp_imap_address(x->envelope->to,  y->envelope->to);
+}
+
+static ImapResponse
+imap_mbox_sort_msgno_client(ImapMboxHandle *handle, ImapSortKey key,
+                            int ascending, unsigned int *msgno, unsigned cnt)
+{
+  /* FIXME: we possibly do not need content type but I am not sure
+   * about all the impliciations... */
+  static const ImapFetchType fetch_type 
+    = IMFETCH_UID | IMFETCH_ENV | IMFETCH_RFC822SIZE | IMFETCH_CONTENT_TYPE;
+  int (*sortfun)(const void *a, const void *b);
+    
+  unsigned i, fetch_cnt, *seqno_to_fetch;
+  struct SortItem *sort_items;
+
+  if(key == IMSO_MSGNO) {
+    g_warning("IMSO_MSGNO not yet implemented.");
+    return IMR_NO;
+  }
+  seqno_to_fetch = g_new(unsigned, cnt);
+  for(i=fetch_cnt=0; i<cnt; i++) {
+    ImapMessage *imsg = imap_mbox_handle_get_msg(handle, msgno[i]);
+    if(!imsg || !imsg->envelope)
+      seqno_to_fetch[fetch_cnt++] = msgno[i];
+  }
+
+  if(fetch_cnt>0) {
+    ImapResponse rc;
+    qsort(seqno_to_fetch, fetch_cnt, sizeof(unsigned), comp_unsigned);
+    printf("Should the client side sorting code "
+           "be sorry about your bandwidth usage?\n");
+    rc = imap_mbox_handle_fetch_set(handle, seqno_to_fetch,
+                                    fetch_cnt, fetch_type);
+    if(rc != IMR_OK)
+      return rc;
+  }
+  g_free(seqno_to_fetch);
+
+  sort_items = g_new(struct SortItem, cnt);
+  for(i=0; i<cnt; i++) {
+    sort_items[i].msg = imap_mbox_handle_get_msg(handle, msgno[i]);
+    sort_items[i].no  = msgno[i];
+  }
+  switch(key) {
+  default:
+  case IMSO_ARRIVAL: sortfun = comp_arrival; break;
+  case IMSO_CC:      sortfun = comp_cc;      break;
+  case IMSO_DATE:    sortfun = comp_date;    break;
+  case IMSO_FROM:    sortfun = comp_from;    break;
+  case IMSO_SIZE:    sortfun = comp_size;    break;
+  case IMSO_SUBJECT: sortfun = comp_subject; break;
+  case IMSO_TO:      sortfun = comp_to;      break;
+  }
+  qsort(sort_items, cnt, sizeof(struct SortItem), sortfun);
+  if(ascending)
+    for(i=0; i<cnt; i++)
+      msgno[i] = sort_items[i].no;
+  else
+    for(i=0; i<cnt; i++)
+      msgno[i] = sort_items[cnt-i-1].no;
+
+
+  g_free(sort_items);
+  return IMR_OK;
+}
+
+ImapResponse
+imap_mbox_sort_msgno(ImapMboxHandle *handle, ImapSortKey key,
+                     int ascending, unsigned int *msgno, unsigned cnt)
+{
+  IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
+  if(imap_mbox_handle_can_do(handle, IMCAP_SORT)) 
+    return imap_mbox_sort_msgno_srv(handle, key, ascending, msgno, cnt);
+  else {
+    return 
+      handle->enable_client_sort
+      ? imap_mbox_sort_msgno_client(handle, key, ascending, msgno, cnt)
+      : IMR_NO;
+  }
+}
+
 static void
 append_no(ImapMboxHandle *handle, unsigned seqno, void *arg)
 {
   mbox_view_append_no(&handle->mbox_view, seqno);
 }
 
+/** selects a subset of messages specified by given filter and sorts
+    it according to specified order. There are four cases possible,
+    depending whether filtering needs to be done and whether the SORT
+    extension is available.
+    CASE 1a: filtered msgno sort.
+    CASE 1b: full     msgno sort.
+    CASE 2a: other sort when SORT extension is present.
+    CASE 2b: other sort without SORT extension.
+*/
 ImapResponse
 imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
                       ImapSearchKey *filter)
@@ -1281,12 +1498,12 @@ imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
 
   IMAP_REQUIRED_STATE1(handle, IMHS_SELECTED, IMR_BAD);
   if(key == IMSO_MSGNO) {
-    if(filter) {
+    if(filter) { /* CASE 1a */
       handle->mbox_view.entries = 0; /* FIXME: I do not like this! 
                                       * we should not be doing such 
                                       * low level manipulations here */
       rc = imap_search_exec(handle, FALSE, filter, append_no, NULL);
-    } else {
+    } else { /* CASE 1b */
       if(handle->thread_root)
         g_node_destroy(handle->thread_root);
       handle->thread_root = g_node_new(NULL);
@@ -1301,36 +1518,55 @@ imap_mbox_sort_filter(ImapMboxHandle *handle, ImapSortKey key, int ascending,
       }
       return IMR_OK;
     }
-  } else { /* Nontrivial (ie. not over msgno) sort requires an extension */
-    unsigned cmdno;
-    int can_do_literals =
-      imap_mbox_handle_can_do(handle, IMCAP_LITERAL);
-    ImapCmdTag tag;
+  } else { 
+    if(imap_mbox_handle_can_do(handle, IMCAP_SORT)) { /* CASE 2a */
+      unsigned cmdno;
+      int can_do_literals =
+        imap_mbox_handle_can_do(handle, IMCAP_LITERAL);
+      ImapCmdTag tag;
 
-    if(!imap_mbox_handle_can_do(handle, IMCAP_SORT)) 
-      return IMR_NO;
-    
-    cmdno =  imap_make_tag(tag);
-    keystr = sort_code_to_string(key);
-    imap_handle_idle_disable(handle);
-    sio_printf(handle->sio, "%s SORT (%s%s) UTF-8 ", tag,
-               ascending ? "" : "REVERSE ", keystr);
-    
-    handle->mbox_view.entries = 0; /* FIXME: I do not like this! 
-                                    * we should not be doing such 
-                                    * low level manipulations here */
-    if(!filter)
-      sio_write(handle->sio, "ALL", 3);
-    else {
-      if( (rc=imap_write_key(handle, filter, cmdno, can_do_literals)) !=IMR_OK)
+      cmdno =  imap_make_tag(tag);
+      keystr = sort_code_to_string(key);
+      imap_handle_idle_disable(handle);
+      sio_printf(handle->sio, "%s SORT (%s%s) UTF-8 ", tag,
+                 ascending ? "" : "REVERSE ", keystr);
+
+      handle->mbox_view.entries = 0; /* FIXME: I do not like this! 
+                                      * we should not be doing such 
+                                      * low level manipulations here */
+      if(!filter)
+        sio_write(handle->sio, "ALL", 3);
+      else {
+        if( (rc=imap_write_key(handle, filter, cmdno, can_do_literals))
+            != IMR_OK)
+          return rc;
+      }
+      sio_write(handle->sio, "\r\n", 2);
+      imap_handle_idle_enable(handle, 30);
+      imap_handle_flush(handle);
+      do
+        rc = imap_cmd_step(handle, cmdno);
+      while(rc == IMR_UNTAGGED);
+    } else {                                           /* CASE 2b */
+      /* try client-side sorting... */
+      if(!handle->enable_client_sort)
+        return IMR_NO;
+      handle->mbox_view.entries = 0; /* FIXME: I do not like this! 
+                                      * we should not be doing such 
+                                      * low level manipulations here */
+      if(filter)
+        rc = imap_search_exec(handle, FALSE, filter, append_no, NULL);
+      else {
+        rc = IMR_OK;
+        for(i=0; i<handle->exists; i++)
+          mbox_view_append_no(&handle->mbox_view, i+1);
+      }
+      if(rc != IMR_OK)
         return rc;
+      rc = imap_mbox_sort_msgno_client(handle, key, ascending,
+                                       handle->mbox_view.arr,
+                                       handle->mbox_view.entries);
     }
-    sio_write(handle->sio, "\r\n", 2);
-    imap_handle_idle_enable(handle, 30);
-    imap_handle_flush(handle);
-    do
-      rc = imap_cmd_step(handle, cmdno);
-    while(rc == IMR_UNTAGGED);
   }
   if(rc == IMR_OK) {
     if(handle->thread_root)
