@@ -153,7 +153,9 @@ imap_mbox_handle_init(ImapMboxHandle *handle)
   handle->info_cb  = NULL;
   handle->info_arg = NULL;
   handle->enable_anonymous = 0;
+  handle->enable_client_sort = 0;
   handle->enable_binary    = 0;
+  handle->enable_idle      = 1;
   mbox_view_init(&handle->mbox_view);
 #if defined(BALSA_USE_THREADS)
   pthread_mutex_init(&handle->mutex, NULL);
@@ -224,7 +226,9 @@ imap_handle_set_option(ImapMboxHandle *h, ImapOption opt, gboolean state)
 {
   switch(opt) {
   case IMAP_OPT_ANONYMOUS: h->enable_anonymous = !!state; break;
+  case IMAP_OPT_CLIENT_SORT: h->enable_client_sort = !!state; break;
   case IMAP_OPT_BINARY:    h->enable_binary    = !!state; break;
+  case IMAP_OPT_IDLE:      h->enable_idle      = !!state; break;
   default: g_warning("imap_set_option: invalid option\n");
   }
 }
@@ -444,7 +448,7 @@ imap_cmd_issue(ImapMboxHandle* h, const char* cmd)
 gboolean
 imap_handle_idle_enable(ImapMboxHandle *h, int seconds)
 {
-  if( !imap_mbox_handle_can_do(h, IMCAP_IDLE))
+  if( !h->enable_idle || !imap_mbox_handle_can_do(h, IMCAP_IDLE))
     return FALSE;
   if(h->idle_issued) {
     fprintf(stderr, "IDLE already enabled\n");
@@ -473,6 +477,12 @@ imap_handle_idle_disable(ImapMboxHandle *h)
   return TRUE;
 }
 
+gboolean
+imap_handle_op_cancelled(ImapMboxHandle *h)
+{
+  return h->op_cancelled;
+}
+
 void
 imap_handle_disconnect(ImapMboxHandle *h)
 {
@@ -632,6 +642,23 @@ imap_socket_open(const char* host, const char *def_port)
   return fd; /* FIXME: provide more info why the connection failed. */
 }
 
+static int
+imap_timeout_cb(void *arg)
+{
+  ImapMboxHandle *h = (ImapMboxHandle*)arg;
+  int ok = 1;
+
+  if(h->user_cb) {
+    h->user_cb(IME_TIMEOUT, h->user_arg, &ok);
+    if(ok) {
+      h->op_cancelled = TRUE;
+      imap_handle_disconnect(h);
+    }
+  }
+
+  return ok;
+}
+
 static ImapResult
 imap_mbox_connect(ImapMboxHandle* handle)
 {
@@ -640,6 +667,7 @@ imap_mbox_connect(ImapMboxHandle* handle)
   const char *service = "imap";
 
   /* reset some handle status */
+  handle->op_cancelled = FALSE;
   handle->has_capabilities = FALSE;
   handle->can_fetch_body = TRUE;
   handle->idle_issued = FALSE;
@@ -662,8 +690,10 @@ imap_mbox_connect(ImapMboxHandle* handle)
     close(handle->sd);
     return IMAP_NOMEM;
   }
-  if(handle->timeout>0)
+  if(handle->timeout>0) {
     sio_set_timeout(handle->sio, handle->timeout);
+    sio_set_timeoutcb(handle->sio, imap_timeout_cb, handle);
+  }
 #ifdef USE_TLS
   if(handle->over_ssl) {
     SSL *ssl = imap_create_ssl();
@@ -731,11 +761,7 @@ imap_get_atom(struct siobuf *sio, char* atom, size_t len)
   for(i=0; i<len-1 && (c=sio_getc(sio)) >=0 && IS_ATOM_CHAR(c); i++)
     atom[i] = c;
 
-  if(i<len-1) {
-    if (c < 0)
-      return c;
-    atom[i] = '\0';
-  } else atom[i+1] = '\0'; /* too long tag?  */
+  atom[i] = '\0';
   return c;
 }
 
@@ -994,6 +1020,7 @@ imap_mbox_set_view(ImapMboxHandle *h, ImapMsgFlag fl, gboolean state)
 {
   char *flag;
   gchar * cmd;
+  const gchar *cmd_prefix;
   void *arg;
   ImapSearchCb cb;
   ImapResponse rc;
@@ -1002,6 +1029,11 @@ imap_mbox_set_view(ImapMboxHandle *h, ImapMsgFlag fl, gboolean state)
   if(fl==0)
     return 1;
 
+  if(imap_mbox_handle_can_do(h, IMCAP_ESEARCH))
+    cmd_prefix = "SEARCH RETURN (ALL) ";
+  else
+    cmd_prefix = "SEARCH ALL ";
+
   switch(fl) {
   case IMSGF_SEEN:     flag = "SEEN"; break;
   case IMSGF_ANSWERED: flag = "ANSWERED"; break;
@@ -1015,7 +1047,7 @@ imap_mbox_set_view(ImapMboxHandle *h, ImapMsgFlag fl, gboolean state)
   arg = h->search_arg; h->search_arg = NULL;
   g_free(h->mbox_view.filter_str);
   h->mbox_view.filter_str = g_strconcat(state ? "" : "UN", flag, NULL);
-  cmd = g_strconcat("SEARCH ALL ", h->mbox_view.filter_str, NULL);
+  cmd = g_strconcat(cmd_prefix, h->mbox_view.filter_str, NULL);
   rc = imap_cmd_exec(h, cmd);
   g_free(cmd);
   h->search_cb = cb; h->search_arg = arg;
@@ -1532,7 +1564,10 @@ imap_cmd_step(ImapMboxHandle* handle, unsigned lastcmd)
       g_free(ci);
     }
   }
-  return lastcmd == cmdno ? rc : IMR_UNTAGGED;
+  if (handle->state == IMHS_DISCONNECTED)
+    return IMR_SEVERED;
+  else
+    return lastcmd == cmdno ? rc : IMR_UNTAGGED;
 }
 
 /* imap_cmd_exec: 
@@ -1614,24 +1649,26 @@ lbi_strerror(ImapResult rc)
   }
 }
 
-static char*
+static GString*
 imap_get_string_with_lookahead(struct siobuf* sio, int c)
-{ /* string */
-  char *res;
-  
+{ /* string */  
+  GString *res = NULL;
   if(c=='"') { /* quoted */
-    GString *str = g_string_new("");
+    res = g_string_new("");
     while( (c=sio_getc(sio)) != '"') {
       if(c== '\\')
         c = sio_getc(sio);
-      g_string_append_c(str, c);
+      g_string_append_c(res, c);
     }
-    res = g_string_free(str, FALSE);
   } else { /* this MUST be literal */
     char buf[15];
     int len;
-    if(c!='{')
-      return NULL;
+    if(c=='~') /* BINARY extension literal8 indicator */
+      c = sio_getc(sio);
+    if(c!='{') {
+      g_string_free(res, TRUE);
+      return NULL; /* ERROR */
+    }
 
     c = imap_get_atom(sio, buf, sizeof(buf));
     len = strlen(buf); 
@@ -1640,9 +1677,10 @@ imap_get_string_with_lookahead(struct siobuf* sio, int c)
     len = atoi(buf);
     if( c != 0x0d) { printf("lit1:%d\n",c); return NULL;}
     if( (c=sio_getc(sio)) != 0x0a) { printf("lit1:%d\n",c); return NULL;}
-    res = g_malloc(len+1);
-    if(len>0) sio_read(sio, res, len);
-    res[len] = '\0';
+    res = g_string_sized_new(len+1);
+    if(len>0) sio_read(sio, res->str, len);
+    res->len = len;
+    res->str[len] = '\0';
   }
   return res;
 }
@@ -1651,7 +1689,8 @@ imap_get_string_with_lookahead(struct siobuf* sio, int c)
 static char*
 imap_get_string(struct siobuf* sio)
 {
-  return imap_get_string_with_lookahead(sio, sio_getc(sio));
+  GString * s = imap_get_string_with_lookahead(sio, sio_getc(sio));
+  return g_string_free(s, FALSE);
 }
 
 static gboolean
@@ -1669,7 +1708,7 @@ imap_get_nstring(struct siobuf* sio)
   if(toupper(c)=='N') { /* nil */
     sio_getc(sio); sio_getc(sio); /* ignore i and l */
     return NULL;
-  } else return imap_get_string_with_lookahead(sio, c);
+  } else return g_string_free(imap_get_string_with_lookahead(sio, c), FALSE);
 }
 
 /* see the spec for the definition of astring */
@@ -1689,12 +1728,24 @@ imap_get_astring(struct siobuf *sio, int* lookahead)
     res = g_string_free(str, FALSE);
     *lookahead = c;
   } else {
-    res = imap_get_string_with_lookahead(sio, c);
+    res = g_string_free(imap_get_string_with_lookahead(sio, c), FALSE);
     *lookahead = sio_getc(sio);
   }
   return res;
 }
 
+/* nstring / literal8 as in the BINARY extension */
+static GString*
+imap_get_binary_string(struct siobuf *sio)
+{
+  int c = sio_getc(sio);
+  if(toupper(c)=='N') { /* nil */
+    sio_getc(sio); sio_getc(sio); /* ignore i and l */
+    return g_string_new("");
+  } else
+    return imap_get_string_with_lookahead(sio, c);
+}
+
 /* this file contains all the response handlers as defined in
    draft-crspin-imapv-20.txt. 
   
@@ -1731,7 +1782,8 @@ ir_capability_data(ImapMboxHandle *handle)
     "AUTH=ANONYMOUS", "AUTH=CRAM-MD5", "AUTH=GSSAPI", "AUTH=PLAIN",
     "STARTTLS", "LOGINDISABLED", "SORT",
     "THREAD=ORDEREDSUBJECT", "THREAD=REFERENCES",
-    "UNSELECT", "SCAN", "CHILDREN", "LITERAL+", "IDLE", "SASL-IR"
+    "UNSELECT", "SCAN", "CHILDREN", "LITERAL+", "IDLE", "SASL-IR",
+    "BINARY", "ESEARCH"
   };
   unsigned x;
   int c;
@@ -1780,7 +1832,7 @@ ir_resp_text_code(ImapMboxHandle *h)
   case 8: imap_get_atom(h->sio, buf, sizeof(buf)); h->uidnext=atoi(buf); break;
   case 9: imap_get_atom(h->sio, buf, sizeof(buf)); h->uidval =atoi(buf); break;
   case 10:imap_get_atom(h->sio, buf, sizeof(buf)); h->unseen =atoi(buf); break;
-  default: while( (c=sio_getc(h->sio)) != EOF && c != ']') ; break;
+  default: while( c != ']' && (c=sio_getc(h->sio)) != EOF) ; break;
   }
   return c != EOF ? rc : IMR_SEVERED;
 }
@@ -1795,15 +1847,18 @@ ir_ok(ImapMboxHandle *h)
     /* look for information response codes here: section 7.1 of the draft */
     rc = ir_resp_text_code(h);
     if(sio_getc(h->sio) != ' ') rc = IMR_PROTOCOL;
-    sio_gets(h->sio, line, sizeof(line));
+    if (sio_gets(h->sio, line, sizeof(line)) == NULL)
+      rc = IMR_SEVERED;
   } else {
     line[0] = c;
-    sio_gets(h->sio, line+1, sizeof(line)-1);
-    rc = IMR_OK;
+    if (sio_gets(h->sio, line+1, sizeof(line)-1) == NULL)
+      rc = IMR_SEVERED;
+    else 
+      rc = IMR_OK;
   }
   if(rc == IMR_PARSE)
     rc = IMR_OK;
-  else if( (l=strlen(line))>0) { 
+  else if (rc != IMR_SEVERED && (l=strlen(line))>0 ) {
     line[l-2] = '\0'; 
     imap_mbox_handle_set_msg(h, line);
     if(h->info_cb)
@@ -1812,7 +1867,7 @@ ir_ok(ImapMboxHandle *h)
       printf("INFO : '%s'\n", line);
     rc = IMR_OK; /* in case it was IMR_ALERT */
   }
-  return IMR_OK;
+  return rc;
 }
 
 static ImapResponse
@@ -2000,6 +2055,99 @@ ir_status(ImapMboxHandle *h)
   return ir_check_crlf(h, sio_getc(h->sio));
 }
 
+/** Process ESEARCH response. Consult RFC4466 and RFC4731 before
+   modification.  */
+static ImapResponse
+ir_esearch(ImapMboxHandle *h)
+{
+  char atom[LONG_STRING];
+  int uid_data = 0;
+  int c = sio_getc(h->sio);
+  if(c == '(') { /* search correlator */
+    gchar *str;
+    c = imap_get_atom(h->sio, atom, sizeof(atom));
+    if(c == EOF) return IMR_SEVERED;
+    if(strcasecmp(atom, "TAG")) { /* TAG is the only acceptable response here! */
+      printf("ESearch expected TAG encountered %s\n", atom);
+      return IMR_PROTOCOL; 
+    }
+    if(c != ' ')
+      return IMR_PROTOCOL;
+    str = imap_get_string(h->sio);
+    /* printf("ESearch response for tag %s\n", str); */
+    g_free(str);
+    if( (c = sio_getc(h->sio)) != ')') {
+      return c == EOF ? IMR_SEVERED : IMR_PROTOCOL;
+    }
+    c = sio_getc(h->sio);
+  }
+  if(c == EOF) return IMR_SEVERED;  
+  if (c == '\r' || c == '\n')
+    return ir_check_crlf(h, c);
+  /* Now, an atom has to follow */
+  c = imap_get_atom(h->sio, atom, sizeof(atom));
+
+  if(strcasecmp(atom, "UID") == 0) {
+    uid_data = 1;
+    c = imap_get_atom(h->sio, atom, sizeof(atom));
+  }
+  if(c == EOF) return IMR_SEVERED;
+
+  while(c == ' ') {
+    static const unsigned LENGTH_OF_LARGEST_UNSIGNED = 10;
+    char value[30];
+    gchar *p;
+    int offset = 0;
+    /* atom contains search-modifier-name, time to fetch
+       search-return-value. In ESEARCH, it is always an atom, so we
+       cut the corners here. We get values in chunks.  The chunk size
+       is pretty arbitrary as long as it can fit two largest possible
+       32-bit unsigned numbers and a colon. */
+    do {
+      size_t value_len;
+      if(c != '\r' && c != '\n') /* Dont try to read beyond the line. */
+        c = imap_get_atom(h->sio, value + offset, sizeof(value)-offset);
+      value_len = strlen(value);
+
+      if(h->search_cb) {
+        unsigned seq, seq1;
+
+        for(p=value; *p &&
+              (unsigned)(p-value) <=
+              sizeof(value)-(2*LENGTH_OF_LARGEST_UNSIGNED+1); ) {
+          if(sscanf(p, "%u", &seq) != 1)
+            return IMR_PROTOCOL;
+          while(*p && isdigit(*p)) p++;
+          if( *p == ':') {
+            p++;
+            if(sscanf(p, "%u", &seq1) != 1)
+              return IMR_PROTOCOL;
+          } else seq1 = seq;
+
+          for(; seq<=seq1; seq++) {
+            h->search_cb(h, seq, h->search_arg);
+          }
+          while(*p && isdigit(*p)) p++;
+          if(*p == ',') p++;
+        } /* End of for */
+	
+	  /* Reuse what's left. */
+        if(*p) {
+          offset = value_len - (p-value);
+          memmove(value, p, offset+1);
+        } else offset = 0;
+
+      } /* End of if(h->search_cb) */
+
+    }  while (isdigit(c)|| offset);
+
+    if(c == ' ')
+      c = imap_get_atom(h->sio, value, sizeof(value));
+  }
+
+  return ir_check_crlf(h, c);
+}
+
 static ImapResponse
 ir_search(ImapMboxHandle *h)
 {
@@ -2999,15 +3147,17 @@ static ImapResponse
 ir_body_section(struct siobuf *sio, unsigned seqno,
 		ImapFetchBodyCb body_cb, void *arg)
 {
-  char buf[80], *str;
+  char buf[80];
+  GString *bs;
   int c = imap_get_atom(sio, buf, sizeof(buf));
+
   if(c != ']') { puts("] expected"); return IMR_PROTOCOL; }
   if(sio_getc(sio) != ' ') { puts("space expected"); return IMR_PROTOCOL;}
-  str = imap_get_nstring(sio);
-  if(str) {
-    if(body_cb)
-      body_cb(seqno, str, strlen(str), arg);
-    g_free(str);
+  bs = imap_get_binary_string(sio);
+  if(bs) {
+    if(bs->str && body_cb)
+      body_cb(seqno, bs->str, bs->len, arg);
+    g_string_free(bs, TRUE);
   }
   return IMR_OK;
 }
@@ -3130,6 +3280,7 @@ ir_fetch_seq(ImapMboxHandle *h, unsigned seqno)
     { "RFC822.HEADER", ir_msg_att_rfc822_header }, 
     { "RFC822.TEXT",   ir_msg_att_rfc822_text }, 
     { "RFC822.SIZE",   ir_msg_att_rfc822_size }, 
+    { "BINARY",        ir_msg_att_body }, 
     { "BODY",          ir_msg_att_body }, 
     { "BODYSTRUCTURE", ir_msg_att_bodystructure }, 
     { "UID",           ir_msg_att_uid }
@@ -3150,7 +3301,7 @@ ir_fetch_seq(ImapMboxHandle *h, unsigned seqno)
     }
     atom[i] = '\0';
     for(i=0; i<ELEMENTS(msg_att); i++) {
-      if(strcmp(atom, msg_att[i].name) == 0) {
+      if(strcasecmp(atom, msg_att[i].name) == 0) {
         if( (rc=msg_att[i].handler(h, c, seqno)) != IMR_OK)
           return rc;
         break;
@@ -3276,11 +3427,11 @@ static const struct {
   { "LIST",       4, ir_list },
   { "LSUB",       4, ir_lsub },
   { "STATUS",     6, ir_status },
+  { "ESEARCH",    7, ir_esearch },
   { "SEARCH",     6, ir_search },
   { "SORT",       4, ir_sort   },
   { "THREAD",     6, ir_thread },
   { "FLAGS",      5, ir_flags  },
-  /* FIXME Is there an unnumbered FETCH response? */
   { "FETCH",      5, ir_fetch  }
 };
 static const struct {
