@@ -125,10 +125,11 @@ void
 pop_set_option(PopHandle *pop, PopOption opt, gboolean state)
 {
   switch(opt) {
-  case IMAP_POP_OPT_DISABLE_APOP: pop->disable_apop = state; break;
-  case IMAP_POP_OPT_FILTER_CR   : pop->filter_cr    = state; break;
-  case IMAP_POP_OPT_OVER_SSL    : pop->over_ssl     = state; break;
-  case IMAP_POP_OPT_PIPELINE    : pop->enable_pipe  = state; break;
+  case IMAP_POP_OPT_DISABLE_APOP: pop->disable_apop = !!state; break;
+  case IMAP_POP_OPT_FILTER_CR   : pop->filter_cr    = !!state; break;
+  case IMAP_POP_OPT_OVER_SSL    : pop->over_ssl     = !!state; break;
+  case IMAP_POP_OPT_PIPELINE    : pop->enable_pipe  = !!state; break;
+  default: g_warning("pop_set_option: invalid option\n");
   }
 }
 
@@ -175,7 +176,7 @@ pop_check_status(PopHandle *pop, GError **err)
     pop->state = IMHS_DISCONNECTED;
     sio_detach(pop->sio); pop->sio = NULL; close(pop->sd);
     g_set_error(err, IMAP_ERROR, IMAP_POP_SEVERED_ERROR,
-                "Connection severed");
+                "POP3 Connection severed");
     return FALSE;
   }
      
@@ -238,7 +239,7 @@ pop_get_capa(PopHandle *pop, GError **err)
     pop->state = IMHS_DISCONNECTED;
     sio_detach(pop->sio); pop->sio = NULL; close(pop->sd);
     g_set_error(err, IMAP_ERROR, IMAP_POP_SEVERED_ERROR,
-                "Connection severed");
+                "POP3 Connection severed");
     return FALSE;
   }
   return TRUE;
@@ -254,23 +255,36 @@ pop_auth_cram(PopHandle *pop, const char *greeting, GError **err)
   return IMAP_AUTH_UNAVAIL;
 }
 
-/* getApopStamp:
-   Get the Server Timestamp for APOP authentication -kabir 
-   return TRUE on success.
-*/
+#define IS_ALLOWED_CHAR(s) ( (s) >=32 && (s) <=126  && (s) !='>')
+/** Get the Timestamp as sent by server for APOP authentication
+   verifying that the syntax is correct. Currently we check only
+   roughly the syntax for unallowed characters.  Return TRUE on if
+   stamp is found and validates.
 
+   @param stamp points to a buffer of length POP_LINE_LEN
+*/
 static gboolean
 get_apop_stamp(const char *greeting, char *stamp)
 {
-  char *start, *finish;
+  const char *start;
 
   start = strchr(greeting, '<');
   if (start) {
-    finish = strchr(start, '>');
-    if (finish) {
-      strncpy(stamp, start, finish - start + 1);
-      return TRUE;
-    }
+    int pos = 0;
+    /* '<' */
+    stamp[pos++] = *start++;
+
+    while( pos < POP_LINE_LEN && IS_ALLOWED_CHAR(*start))
+      stamp[pos++] = *start++;
+    if(*start != '>' || !(pos < POP_LINE_LEN) )
+      { return FALSE; }
+
+    /* '>' */
+    stamp[pos++] = *start++;
+    if( !(pos < POP_LINE_LEN) )
+      { return FALSE; }
+    stamp[pos] = '\0';
+    return TRUE;
   } 
   return FALSE;
 }
@@ -418,9 +432,11 @@ pop_stls(PopHandle *pop, GError **err)
     pop->tls_enabled = 1;
     return TRUE;
   } else {
-    SSL_free(ssl);
+    /* this will destroy ssl, too */
     sio_detach(pop->sio); pop->sio = NULL; close(pop->sd);
     pop->state = IMHS_DISCONNECTED;
+    g_set_error(err, IMAP_ERROR, IMAP_POP_SEVERED_ERROR,
+                "TLS setup failed");
     return FALSE;
   }
 }
@@ -523,7 +539,8 @@ pop_connect(PopHandle *pop, const char *host, GError **err)
   
 #ifdef USE_TLS
   if(pop->tls_mode != IMAP_TLS_DISABLED && pop_can_do(pop, POP_CAP_STLS)) {
-    pop_stls(pop, err); /* Try... */
+    if(!pop_stls(pop, err)) /* TLS negotiation attempted.. */
+      return FALSE;         /* .. but failed. */
   }
 #endif
   if(pop->tls_mode == IMAP_TLS_REQUIRED && 
@@ -580,6 +597,11 @@ pop_get_uid(PopHandle *pop, unsigned msgno, GError **err)
   if(!pop->uids) {
     char line[POP_LINE_LEN];
     unsigned curr_msgno = 1;
+    if(!pop->sio) {
+      g_set_error(err, IMAP_ERROR, IMAP_POP_SEVERED_ERROR,
+                  "POP3 Connection severed");
+      return NULL;
+    }
     if(!pop_exec(pop, "UIDL\r\n", err))
       return NULL;
     pop->uids = g_ptr_array_sized_new(pop->msg_cnt);
