@@ -33,10 +33,25 @@
 #include <stdarg.h>
 #include <unistd.h>
 
+#ifdef HAVE_NOTIFY
+#include <libnotify/notify.h>
+#endif
+
 #if ENABLE_LDAP
 #include <ldap.h>
 #endif
 
+#if HAVE_COMPFACE
+#include <compface.h>
+#endif                          /* HAVE_COMPFACE */
+
+#if HAVE_GTKSOURCEVIEW
+#include <gtksourceview/gtksourceview.h>
+#include <gtksourceview/gtksourcebuffer.h>
+#include <gtksourceview/gtksourcetag.h>
+#include <gtksourceview/gtksourcetagstyle.h>
+#endif
+
 #include "libbalsa.h"
 #include "misc.h"
 #include "i18n.h"
@@ -46,12 +61,6 @@ static pthread_t main_thread_id;
 static pthread_t libbalsa_threads_id;
 #endif
 
-#define POP_SERVER "pop"
-#define IMAP_SERVER "mx"
-#define LDAP_SERVER "ldap"
-
-
-static gchar *qualified_hostname(const char *name);
 
 void
 libbalsa_message(const char *fmt, ...)
@@ -69,6 +78,9 @@ libbalsa_init(LibBalsaInformationFunc information_callback)
 {
     struct utsname utsname;
 
+#ifdef HAVE_NOTIFY
+    notify_init("Basics");
+#endif
 
 #ifdef BALSA_USE_THREADS
     if (!g_thread_supported()) {
@@ -210,123 +222,15 @@ libbalsa_guess_mail_spool(void)
     return g_strconcat(g_get_home_dir(), "/mailbox", NULL);
 }
 
-/* Some more "guess" functions symmetric to libbalsa_guess_mail_spool()... */
-
-
-static gchar *qualified_hostname(const char *name)
-{
-    gchar *domain=libbalsa_get_domainname();
-
-    if(domain) {
-	gchar *host=g_strdup_printf("%s.%s", name, domain);
-	
-	g_free(domain);
-
-	return host;
-    } else
-	return g_strdup(name);
-}
-
-
-gchar *libbalsa_guess_pop_server()
-{
-    return qualified_hostname(POP_SERVER);
-}
-
-gchar *libbalsa_guess_imap_server()
-{
-    return qualified_hostname(IMAP_SERVER);
-}
-
-gchar *libbalsa_guess_ldap_server()
-{
-    return qualified_hostname(LDAP_SERVER);
-}
-
-gchar *libbalsa_guess_imap_inbox()
-{
-    gchar *server = libbalsa_guess_imap_server();
-
-    if(server) {
-	gchar *url = g_strdup_printf("imap://%s/INBOX", server);
-	
-	g_free(server);
-
-	return url;
-    }
-
-    return NULL;
-}
-
-gchar *libbalsa_guess_ldap_base()
-{
-    gchar *server = libbalsa_guess_ldap_server();
-
-    /* Note: Assumes base dn is "o=<domain name>". Somewhat speculative... */
-    if(server) {
-	gchar *base=NULL, *domain;
-
-	if((domain=strchr(server, '.')))
-	   base = g_strdup_printf("o=%s", domain+1);
-	
-	g_free(server);
-
-	return base;
-    }
-    return NULL;
-}
-
-gchar *libbalsa_guess_ldap_name()
-{
-    gchar *base = libbalsa_guess_ldap_base();
-
-    if(base) {
-	gchar *name = strchr(base, '=');
-	gchar *dir_name = g_strdup_printf(_("LDAP Directory for %s"), 
-					  (name?name+1:base));
-	g_free(base);
-
-	return dir_name;
-    } 
-
-    return NULL;
-}
-
-gchar *libbalsa_guess_ldif_file()
-{
-    int i;
-    gchar *ldif;
-
-    static const gchar *guesses[] = {
-	"address.ldif",
-	".address.ldif",
-	"address-book.ldif",
-	".address-book.ldif",
-	".addressbook.ldif",
-	NULL
-    };
-
-    for (i = 0; guesses[i] != NULL; i++) {
-	ldif =  g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, 
-			    guesses[i], NULL);
-	if (g_file_test(ldif, G_FILE_TEST_EXISTS))
-	     return ldif;
-	  
-	g_free(ldif);
-    }
-    return g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, 
-			guesses[i], NULL); /* *** Or NULL */
-    
-}
 
 gboolean libbalsa_ldap_exists(const gchar *server)
 {
 #if ENABLE_LDAP
-    LDAP *ldap = ldap_open(server, LDAP_PORT);
+    LDAP *ldap;
+    ldap_initialize(&ldap, server);
 
     if(ldap) {
-	ldap_unbind(ldap);
-
+	ldap_unbind_ext(ldap, NULL, NULL);
 	return TRUE;
     }
 #endif /* #if ENABLE_LDAP */
@@ -334,14 +238,6 @@ gboolean libbalsa_ldap_exists(const gchar *server)
     return FALSE;
 }
 
-void
-libbalsa_assure_balsa_dir(void)
-{
-    gchar* dir = g_strconcat(g_get_home_dir(), "/.balsa", NULL);
-    mkdir(dir, S_IRUSR|S_IWUSR|S_IXUSR);
-    g_free(dir);
-}
-
 gchar*
 libbalsa_date_to_utf8(const time_t *date, const gchar *date_string)
 {
@@ -376,6 +272,73 @@ libbalsa_get_icon_from_flags(LibBalsaMessageFlag flags)
 }
 
 
+#ifdef BALSA_USE_THREADS
+#include <pthread.h>
+typedef struct {
+    pthread_mutex_t lock;
+    pthread_cond_t condvar;
+    int (*cb)(void *arg);
+    void *arg;
+    int res;
+} AskData;
+
+/* ask_cert_idle:
+   called in MT mode by the main thread.
+ */
+static gboolean
+ask_idle(gpointer data)
+{
+    AskData* ad = (AskData*)data;
+    printf("ask_idle: ENTER %p\n", data);
+    gdk_threads_enter();
+    ad->res = (ad->cb)(ad->arg);
+    gdk_threads_leave();
+    pthread_cond_signal(&ad->condvar);
+    printf("ask_idle: LEAVE %p\n", data);
+    return FALSE;
+}
+
+/* libbalsa_ask_mt:
+   executed with GDK UNLOCKED. see mailbox_imap_open() and
+   imap_dir_cb()/imap_folder_imap_dir().
+*/
+static int
+libbalsa_ask(gboolean (*cb)(void *arg), void *arg)
+{
+    AskData ad;
+
+    if (pthread_self() == main_thread_id) {
+        int ret;
+        printf("Main thread asks the following question.\n");
+        gdk_threads_enter();
+        ret = cb(arg);
+        gdk_threads_leave();
+        return ret;
+    }
+    printf("Side thread asks the following question.\n");
+    pthread_mutex_init(&ad.lock, NULL);
+    pthread_cond_init(&ad.condvar, NULL);
+    ad.cb  = cb;
+    ad.arg = arg;
+
+    pthread_mutex_lock(&ad.lock);
+    pthread_cond_init(&ad.condvar, NULL);
+    g_idle_add(ask_idle, &ad);
+    pthread_cond_wait(&ad.condvar, &ad.lock);
+
+    pthread_cond_destroy(&ad.condvar);
+    pthread_mutex_unlock(&ad.lock);
+    return ad.res;
+}
+#else /* BALSA_USE_THREADS */
+static gboolean
+libbalsa_ask(gboolean (*cb)(void *arg), void *arg)
+{
+    return cb(arg);
+}
+#endif /* BALSA_USE_THREADS */
+
+
 #if defined(USE_SSL)
 #include <openssl/ssl.h>
 #include <openssl/x509.h>
@@ -530,12 +493,19 @@ libbalsa_is_cert_known(X509* cert, long vfy_result)
    TODO: check treading issues.
 
 */
-static int
-ask_cert_real(X509 *cert, const char *explanation)
-{
+struct AskCertData {
+    X509 *certificate;
+    const char *explanation;
+};
 
-    char *part[] =
+static int
+ask_cert_real(void *data)
+{
+    static const char *part[] =
         {"/CN=", "/Email=", "/O=", "/OU=", "/L=", "/ST=", "/C="};
+
+    struct AskCertData *acd = (struct AskCertData*)data;
+    X509 *cert = acd->certificate;
     char buf[256]; /* fingerprint requires EVP_MAX_MD_SIZE*3 */
     char *name = NULL, *c, *valid_from, *valid_until;
     GtkWidget* dialog, *label;
@@ -547,7 +517,7 @@ ask_cert_real(X509 *cert, const char *explanation)
                            "could not be verified.\n"
                            "<b>Reason:</b> %s\n"
                            "<b>This certificate belongs to:</b>\n"),
-                    explanation);
+                    acd->explanation);
 
     name = X509_NAME_oneline(X509_get_subject_name (cert), buf, sizeof (buf));
     for (i = 0; i < ELEMENTS(part); i++) {
@@ -604,62 +574,51 @@ ask_cert_real(X509 *cert, const char *explanation)
     return i;
 }
 
-#ifdef BALSA_USE_THREADS
-#include <pthread.h>
-typedef struct {
-    pthread_cond_t cond;
-    X509 *cert;
-    const char *expl;
-    int res;
-} AskCertData;
-/* ask_cert_idle:
-   called in MT mode by the main thread.
- */
-static gboolean
-ask_cert_idle(gpointer data)
-{
-    AskCertData* acd = (AskCertData*)data;
-    gdk_threads_enter();
-    acd->res = ask_cert_real(acd->cert, acd->expl);
-    gdk_threads_leave();
-    pthread_cond_signal(&acd->cond);
-    return FALSE;
-}
-/* libmutt_ask_for_cert_acceptance:
-   executed with GDK UNLOCKED. see mailbox_imap_open() and
-   imap_dir_cb()/imap_folder_imap_dir().
-*/
 static int
 libbalsa_ask_for_cert_acceptance(X509 *cert, const char *explanation)
 {
-    static pthread_mutex_t ask_cert_lock = PTHREAD_MUTEX_INITIALIZER;
-    AskCertData acd;
-
-    if (pthread_self() == main_thread_id)
-	return ask_cert_real(cert, explanation);
-
-    pthread_mutex_lock(&ask_cert_lock);
-    pthread_cond_init(&acd.cond, NULL);
-    acd.cert = cert;
-    acd.expl = explanation;
-    g_idle_add(ask_cert_idle, &acd);
-    pthread_cond_wait(&acd.cond, &ask_cert_lock);
-    
-    pthread_cond_destroy(&acd.cond);
-    pthread_mutex_unlock(&ask_cert_lock);
-    return acd.res;
+    struct AskCertData acd;
+    acd.certificate = cert;
+    acd.explanation = explanation;
+    return libbalsa_ask(ask_cert_real, &acd);
 }
-#else /* BALSA_USE_THREADS */
-static int
-libbalsa_ask_for_cert_acceptance(X509 *cert, const char *explanation)
-{
-    return ask_cert_real(cert, explanation);
-}
-#endif /* BALSA_USE_THREADS */
-
 #endif /* WITH_SSL */
 
 
+static int
+ask_timeout_real(void *data)
+{
+    const char *host = (const char*)data;
+    GtkWidget* dialog;
+    int i;
+
+    dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO,
+                                    GTK_BUTTONS_YES_NO,
+                                    _("Connection to %s timed out. Abort?"),
+                                    host);
+    gtk_window_set_wmclass(GTK_WINDOW(dialog), "timeout_dialog", "Balsa");
+    switch(gtk_dialog_run(GTK_DIALOG(dialog))) {
+    case GTK_RESPONSE_YES: i = 1; break;
+    case GTK_RESPONSE_NO: i = 0; break;
+    default: printf("Unknown response. Defaulting to 'yes'.\n");
+        i = 1;
+    }
+    gtk_widget_destroy(dialog);
+    /* Process some events to let the window disappear:
+     * not really necessary but helps with debugging. */
+   while(gtk_events_pending()) 
+        gtk_main_iteration_do(FALSE);
+    printf("%s returns %d\n", __FUNCTION__, i);
+    return i;
+}
+
+gboolean
+libbalsa_abort_on_timeout(const char *host)
+{
+    return libbalsa_ask(ask_timeout_real, (void*)host) != 0;
+}
+
+
 #ifdef BALSA_USE_THREADS
 pthread_t
 libbalsa_get_main_thread(void)
@@ -687,27 +646,32 @@ void
 libbalsa_lock_mailbox(LibBalsaMailbox * mailbox)
 {
     pthread_t thread_id = pthread_self();
+    gint count = 0;
+
+    if (thread_id == libbalsa_threads_id
+        && thread_id == mailbox->thread_id
+        && mailbox->lock > 0) {
+        /* We already have both locks, so we'll just hold on to both of
+         * them. */
+        ++mailbox->lock;
+#if LIBBALSA_DEBUG_THREADS
+        g_message("Avoided temporary gdk_threads_leave!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+        return;
+    }
+
+    while (thread_id == libbalsa_threads_id) {
+        ++count;
+#if LIBBALSA_DEBUG_THREADS
+        g_message("Temporary gdk_threads_leave!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+        gdk_threads_leave();
+    }
 
     pthread_mutex_lock(&mailbox_lock);
 
-    while (mailbox->lock && mailbox->thread_id != thread_id) {
-        gint count = 0;
-
-        while (thread_id == libbalsa_threads_id) {
-	    ++count;
-#if LIBBALSA_DEBUG_THREADS
-            g_message("Temporary gdk_threads_leave!!!");
-#endif /* LIBBALSA_DEBUG_THREADS */
-            gdk_threads_leave();
-	}
-	pthread_cond_wait(&mailbox_cond, &mailbox_lock);
-        while (--count >= 0) {
-            gdk_threads_enter();
-#if LIBBALSA_DEBUG_THREADS
-            g_message("...and gdk_threads_enter!!!");
-#endif /* LIBBALSA_DEBUG_THREADS */
-        }
-    }
+    while (mailbox->lock && mailbox->thread_id != thread_id)
+        pthread_cond_wait(&mailbox_cond, &mailbox_lock);
 
     /* We'll assume that no-one would destroy a mailbox while we've been
      * trying to lock it. If they have, we have larger problems than
@@ -716,25 +680,35 @@ libbalsa_lock_mailbox(LibBalsaMailbox * mailbox)
     mailbox->thread_id = thread_id;
 
     pthread_mutex_unlock(&mailbox_lock);
+
+    while (--count >= 0) {
+        gdk_threads_enter();
+#if LIBBALSA_DEBUG_THREADS
+        g_message("...and gdk_threads_enter!!!");
+#endif                          /* LIBBALSA_DEBUG_THREADS */
+    }
 }
 
 void
 libbalsa_unlock_mailbox(LibBalsaMailbox * mailbox)
 {
-#if LIBBALSA_DEBUG_THREADS
     pthread_t self;
 
     self = pthread_self();
 
-    if (self != mailbox->thread_id) {
+    pthread_mutex_lock(&mailbox_lock);
+
+    if (mailbox->lock == 0 || self != mailbox->thread_id) {
 	g_warning("Not holding mailbox lock!!!");
+        pthread_mutex_unlock(&mailbox_lock);
 	return;
     }
 
-#endif /* LIBBALSA_DEBUG_THREADS */
-    pthread_mutex_lock(&mailbox_lock);
-    if(!--mailbox->lock)
+    if(--mailbox->lock == 0) {
         pthread_cond_broadcast(&mailbox_cond);
+        mailbox->thread_id = 0;
+    }
+
     pthread_mutex_unlock(&mailbox_lock);
 }
 
@@ -763,12 +737,11 @@ libbalsa_threads_leave(void)
 
     self = pthread_self();
 
-#if LIBBALSA_DEBUG_THREADS
-    if (self != libbalsa_threads_id) {
+    if (libbalsa_threads_lock == 0 || self != libbalsa_threads_id) {
 	g_warning("Not holding gdk lock!!!");
 	return;
     }
-#endif /* LIBBALSA_DEBUG_THREADS */
+
     if (--libbalsa_threads_lock == 0) {
 	if (self != main_thread_id)
 	    gdk_display_flush(gdk_display_get_default());
@@ -791,4 +764,249 @@ libbalsa_threads_destroy(void)
     pthread_mutex_destroy(&libbalsa_threads_mutex);
 }
 
+gboolean
+libbalsa_threads_has_lock(void)
+{
+    return libbalsa_threads_lock > 0
+        && libbalsa_threads_id == pthread_self();
+}
+
 #endif				/* BALSA_USE_THREADS */
+
+/* Initialized by the front end. */
+void (*libbalsa_progress_set_text) (LibBalsaProgress * progress,
+                                    const gchar * text, guint total);
+void (*libbalsa_progress_set_fraction) (LibBalsaProgress * progress,
+                                        gdouble fraction);
+
+/*
+ * Face and X-Face header support.
+ */
+gchar *
+libbalsa_get_header_from_path(const gchar * header, const gchar * path,
+                              gsize * size, GError ** err)
+{
+    gchar *buf, *content;
+    size_t name_len;
+    gchar *p, *q;
+
+    if (!g_file_get_contents(path, &buf, size, err))
+        return NULL;
+
+    content = buf;
+    name_len = strlen(header);
+    if (g_ascii_strncasecmp(content, header, name_len) == 0)
+        /* Skip header and trailing colon: */
+        content += name_len + 1;
+
+    /* Unfold. */
+    for (p = q = content; *p; p++)
+        if (*p != '\r' && *p != '\n')
+            *q++ = *p;
+    *q = '\0';
+
+    content = g_strdup(content);
+    g_free(buf);
+
+    return content;
+}
+
+GtkWidget *
+libbalsa_get_image_from_face_header(const gchar * content, GError ** err)
+{
+    GMimeStream *stream;
+    GMimeStream *stream_filter;
+    GMimeFilter *filter;
+    GByteArray *array;
+    GtkWidget *image = NULL;
+
+    stream = g_mime_stream_mem_new();
+    stream_filter = g_mime_stream_filter_new_with_stream(stream);
+
+    filter = g_mime_filter_basic_new_type(GMIME_FILTER_BASIC_BASE64_DEC);
+    g_mime_stream_filter_add(GMIME_STREAM_FILTER(stream_filter), filter);
+    g_object_unref(filter);
+
+    g_mime_stream_write_string(stream_filter, content);
+    g_object_unref(stream_filter);
+
+    array = GMIME_STREAM_MEM(stream)->buffer;
+    if (array->len == 0)
+        g_set_error(err, LIBBALSA_IMAGE_ERROR,
+                    LIBBALSA_IMAGE_ERROR_NO_DATA, _("No image data"));
+    else {
+        GdkPixbufLoader *loader =
+            gdk_pixbuf_loader_new_with_type("png", NULL);
+
+        gdk_pixbuf_loader_write(loader, array->data, array->len, err);
+        gdk_pixbuf_loader_close(loader, *err ? NULL : err);
+
+        if (!*err)
+            image = gtk_image_new_from_pixbuf(gdk_pixbuf_loader_get_pixbuf
+                                              (loader));
+        g_object_unref(loader);
+    }
+    g_object_unref(stream);
+
+    return image;
+}
+
+
+GQuark
+libbalsa_image_error_quark(void)
+{
+    static GQuark quark = 0;
+    if (quark == 0)
+        quark = g_quark_from_static_string("libbalsa-image-error-quark");
+    return quark;
+}
+
+#if HAVE_COMPFACE
+GtkWidget *
+libbalsa_get_image_from_x_face_header(const gchar * content, GError ** err)
+{
+    gchar buf[2048];
+    GdkPixbuf *pixbuf;
+    guchar *pixels;
+    gint lines;
+    const gchar *p;
+    GtkWidget *image = NULL;
+
+    strncpy(buf, content, sizeof buf - 1);
+
+    switch (uncompface(buf)) {
+    case -1:
+        g_set_error(err, LIBBALSA_IMAGE_ERROR, LIBBALSA_IMAGE_ERROR_FORMAT,
+                    _("Invalid input format"));
+        return image;
+    case -2:
+        g_set_error(err, LIBBALSA_IMAGE_ERROR, LIBBALSA_IMAGE_ERROR_BUFFER,
+                    _("Internal buffer overrun"));
+        return image;
+    }
+
+    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, 48, 48);
+    pixels = gdk_pixbuf_get_pixels(pixbuf);
+
+    p = buf;
+    for (lines = 48; lines > 0; --lines) {
+        guint x[3];
+        gint j, k;
+        guchar *q;
+
+        if (sscanf(p, "%x,%x,%x,", &x[0], &x[1], &x[2]) != 3) {
+            g_set_error(err, LIBBALSA_IMAGE_ERROR,
+                        LIBBALSA_IMAGE_ERROR_BAD_DATA,
+                        /* Translators: please do not translate Face. */
+                        _("Bad X-Face data"));
+            g_object_unref(pixbuf);
+            return image;
+        }
+        for (j = 0, q = pixels; j < 3; j++)
+            for (k = 15; k >= 0; --k){
+                guchar c = x[j] & (1 << k) ? 0x00 : 0xff;
+                *q++ = c;       /* red   */
+                *q++ = c;       /* green */
+                *q++ = c;       /* blue  */
+            }
+        p = strchr(p, '\n') + 1;
+        pixels += gdk_pixbuf_get_rowstride(pixbuf);
+    }
+
+    image = gtk_image_new_from_pixbuf(pixbuf);
+    g_object_unref(pixbuf);
+
+    return image;
+}
+#endif                          /* HAVE_COMPFACE */
+
+#if HAVE_GTKSOURCEVIEW
+GtkWidget *
+libbalsa_source_view_new(gboolean highlight_phrases, GdkColor *q_colour)
+{
+    GtkTextTag * text_tag;
+    GtkSourceTagStyle *tag_style;
+    GtkSourceTagTable *tag_table;
+    GSList *tag_list;
+    GtkSourceBuffer *sbuffer;
+    GtkWidget *sview;
+
+    /* create the tag table */
+    tag_list = NULL;
+    tag_table = gtk_source_tag_table_new();
+
+    /* add highlighting for quoted text if requested */
+    if (q_colour) {
+	int k;
+
+	for (k = 1; k <= 9; k++) {
+	    gchar * tag_id;
+	    const gchar * pattern;
+            gchar *tmp = NULL;
+
+	    tag_id = g_strdup_printf("Quote-%d", k);
+	    if (k == 1)
+		pattern = "^> *($|[^ |>:}#\n])";
+	    else
+		pattern = tmp =
+                    g_strdup_printf("^(> *){%d}($|[^ |>:}#\n])", k);
+	    text_tag = gtk_line_comment_tag_new(tag_id, tag_id, pattern);
+	    g_free(tmp);
+	    g_free(tag_id);
+	    tag_style = gtk_source_tag_style_new();
+	    tag_style->mask = GTK_SOURCE_TAG_STYLE_USE_FOREGROUND;
+	    tag_style->foreground = q_colour[(k - 1) & 1];
+	    gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	    gtk_source_tag_style_free(tag_style);
+	    tag_list = g_slist_prepend(tag_list, text_tag);
+	}
+    }
+
+    /* if requested create the patterns for bold, italic and underline */
+    if (highlight_phrases) {
+	text_tag = gtk_pattern_tag_new("Bold", "Bold",
+				       "(^|[[:space:]])\\*[[:alnum:]][^*\n]*[[:alnum:]]\\*");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->bold = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+
+	text_tag = gtk_pattern_tag_new("Italic", "Italic",
+				       "(^|[[:space:]])/[[:alnum:]][^/\n]*[[:alnum:]]/");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->italic = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+
+	text_tag = gtk_pattern_tag_new("Underline", "Underline",
+				       "(^|[[:space:]])_[[:alnum:]][^_\n]*[[:alnum:]]_");
+	tag_style = gtk_source_tag_style_new();
+	tag_style->underline = TRUE;
+	gtk_source_tag_set_style(GTK_SOURCE_TAG(text_tag), tag_style);
+	gtk_source_tag_style_free(tag_style);
+	tag_list = g_slist_prepend(tag_list, text_tag);
+    }
+
+    /* add tags to the table if present */
+    if (tag_list) {
+	gtk_source_tag_table_add_tags(tag_table, tag_list);
+	g_slist_foreach(tag_list, (GFunc)g_object_unref, NULL);
+	g_slist_free(tag_list);
+    }
+
+    /* create the source buffer */
+    sbuffer = gtk_source_buffer_new(tag_table);
+    g_object_unref(tag_table);
+    gtk_source_buffer_set_highlight(sbuffer, highlight_phrases || q_colour);
+    gtk_source_buffer_set_check_brackets(sbuffer, FALSE);
+
+    /* create & return the source view */
+    sview = gtk_source_view_new_with_buffer(sbuffer);
+    g_object_unref(sbuffer);
+
+    return sview;
+}
+#endif  /* HAVE_GTKSOURCEVIEW */
+
