@@ -136,7 +136,6 @@ typedef enum {
     LB_MAILBOX_STATE_CLOSED,
     LB_MAILBOX_STATE_OPENING,
     LB_MAILBOX_STATE_OPEN,
-    LB_MAILBOX_STATE_TREECLEANING,
     LB_MAILBOX_STATE_CLOSING
 } LibBalsaMailboxState;
 
@@ -245,6 +244,11 @@ struct _LibBalsaMailbox {
 
     /* Whether the tree has been changed since some event. */
     gboolean msg_tree_changed;
+
+#ifdef BALSA_USE_THREADS
+    /* Array of msgnos that need to be displayed. */
+    GArray *msgnos_pending;
+#endif                          /* BALSA_USE_THREADS */
 };
 
 /* Search iter */
@@ -268,8 +272,7 @@ struct _LibBalsaMailboxClass {
     gboolean (*open_mailbox) (LibBalsaMailbox * mailbox, GError **err);
     void (*close_mailbox) (LibBalsaMailbox * mailbox, gboolean expunge);
     LibBalsaMessage *(*get_message) (LibBalsaMailbox * mailbox, guint msgno);
-    void (*prepare_threading)(LibBalsaMailbox *mailbox, guint * msgnos,
-                              guint len);
+    gboolean (*prepare_threading)(LibBalsaMailbox *mailbox, guint start);
     gboolean (*fetch_message_structure)(LibBalsaMailbox *mailbox,
                                         LibBalsaMessage * message,
                                         LibBalsaFetchFlag flags);
@@ -359,9 +362,11 @@ LibBalsaMessage *libbalsa_mailbox_get_message(LibBalsaMailbox * mailbox,
 /** libbalsa_mailbox_prepare_threading() requests prefetching of information
     needed for client-side message threading.
     msgnos are related to currently set view.
+    Returns TRUE if successful; FALSE may mean that the mailbox was
+    closed during the operation.
 */
-void libbalsa_mailbox_prepare_threading(LibBalsaMailbox *mailbox,
-					guint * msgnos, guint len);
+gboolean libbalsa_mailbox_prepare_threading(LibBalsaMailbox * mailbox,
+                                            guint start);
 
 /** libbalsa_mailbox_fetch_message_structure() fetches detailed
     message structure for given message. It can also fetch all RFC822
@@ -458,7 +463,6 @@ gboolean libbalsa_mailbox_messages_move(LibBalsaMailbox * mailbox,
 					GArray * msgnos,
 					LibBalsaMailbox * dest, GError **err);
 
-                               
 /*
  * misc mailbox releated functions
  */
@@ -579,6 +583,11 @@ gboolean libbalsa_mailbox_msgno_find(LibBalsaMailbox * mailbox,
 				     guint seqno,
 				     GtkTreePath ** path,
 				     GtkTreeIter * iter);
+/* Manage message flags */
+gboolean libbalsa_mailbox_msgno_change_flags(LibBalsaMailbox * mailbox,
+                                             guint msgno,
+                                             LibBalsaMessageFlag set,
+                                             LibBalsaMessageFlag clear);
 /* Test message flags */
 gboolean libbalsa_mailbox_msgno_has_flags(LibBalsaMailbox * mailbox,
                                           guint seqno,
@@ -603,7 +612,9 @@ void libbalsa_mailbox_set_encr_icon(GdkPixbuf * pixbuf);
 void libbalsa_mailbox_try_reassemble(LibBalsaMailbox * mailbox,
 				     const gchar * id);
 
-/* Message number arrays */
+/* Message numbers and arrays */
+void libbalsa_mailbox_register_msgno(LibBalsaMailbox * mailbox,
+                                     guint * msgno);
 void libbalsa_mailbox_register_msgnos(LibBalsaMailbox * mailbox,
 				      GArray * msgnos);
 void libbalsa_mailbox_unregister_msgnos(LibBalsaMailbox * mailbox,
@@ -646,7 +657,6 @@ typedef enum {
     LB_MBOX_SIZE_COL,
     LB_MBOX_WEIGHT_COL,
     LB_MBOX_STYLE_COL,
-    LB_MBOX_MESSAGE_COL,
     LB_MBOX_N_COLS
 } LibBalsaMailboxColumn;
 
