@@ -83,7 +83,6 @@ libbalsa_message_get_type()
     return libbalsa_message_type;
 }
 
-int message_cnt = 0;
 static void
 libbalsa_message_init(LibBalsaMessage * message)
 {
@@ -103,7 +102,6 @@ libbalsa_message_init(LibBalsaMessage * message)
 #ifdef HAVE_GPGME
     message->prot_state = LIBBALSA_MSG_PROTECT_NONE;
 #endif
-    /* printf("%p message created.\n", message); */ message_cnt++;
 }
 
 
@@ -180,12 +178,14 @@ libbalsa_message_finalize(GObject * object)
 	message->mime_msg = NULL;
     }
     G_OBJECT_CLASS(parent_class)->finalize(object);
-    /* printf("%p message finalized.\n", message); */  message_cnt--;
 }
 
 static void
 lb_message_headers_extra_destroy(LibBalsaMessageHeaders * headers)
 {
+    if (!headers)
+	return;
+
     FREE_HEADER_LIST(headers->user_hdrs);
     headers->user_hdrs = NULL;
 
@@ -268,25 +268,28 @@ libbalsa_message_body_charset(LibBalsaMessageBody * body)
     return libbalsa_message_body_charset(body->next);
 }
 
+/* UTF-8-aware header cleaning by Albrecht */
 static void
 canonize_header_value(gchar *value)
 {
-#if 0
-    /* there was some unclarity whether we should do it or not. So we
-       do not for now. */
-    int src, dst;
+    gchar *dptr = value;
 
-    /* canonize value: replace multiple spaces, tabs by single space */
-    for(src=dst=0; value[src]; src++) {
-        if( isspace((int)value[src])) {
-            value[dst] = ' ';
-            if(!isspace((int)value[src+1]))
-                dst++;
-        } else
-            value[dst++] = value[src];
+    while (*value) {
+        if (g_unichar_isspace(g_utf8_get_char(value))) {
+            do {
+                value = g_utf8_next_char(value);
+            } while (g_unichar_isspace(g_utf8_get_char(value)));
+            *dptr++ = ' ';
+        } else {
+            gint bytes = g_utf8_next_char(value) - value;
+
+            do {
+                *dptr++ = *value++;
+            } while (--bytes > 0);
+        }
     }
-    value[dst] = '\0';
-#endif
+
+    *dptr = '\0';
 }
 
 /* message_user_hdrs:
@@ -307,32 +310,69 @@ libbalsa_create_hdr_pair(const gchar * name, gchar * value)
     return item;
 }
 
-/** libbalsa_message_find_user_hdr:
-    returns.... list element matching given header.
-*/
-static GList *
-libbalsa_message_find_user_hdr(LibBalsaMessage * message, const gchar * find)
+static GList*
+libbalsa_message_header_get_helper(LibBalsaMessageHeaders* headers,
+                                   const gchar *find)
 {
-    GList* list;
-    LibBalsaMessageHeaders *headers = message->headers;
-    
-    g_return_val_if_fail(headers, NULL);
-    if (!headers->user_hdrs && message->mailbox) 
-        libbalsa_mailbox_set_msg_headers(message->mailbox, message);
-
+    GList *list;
     for (list = headers->user_hdrs; list; list = list->next) {
         const gchar * const *tmp = list->data;
         
         if (g_ascii_strncasecmp(tmp[0], find, strlen(find)) == 0) 
             return list;
     }
-    
     return NULL;
 }
 
+/** libbalsa_message_find_user_hdr:
+    returns.... list element matching given header.
+*/
+static GList *
+libbalsa_message_find_user_hdr(LibBalsaMessage * message, const gchar * find)
+{
+    LibBalsaMessageHeaders *headers = message->headers;
+    
+    g_return_val_if_fail(headers, NULL);
+    if (!headers->user_hdrs && message->mailbox) 
+        libbalsa_mailbox_set_msg_headers(message->mailbox, message);
+
+    return libbalsa_message_header_get_helper(headers, find);
+}
+
 /* 
  * Public user header methods
  */
+const gchar*
+libbalsa_message_header_get_one(LibBalsaMessageHeaders* headers,
+                                const gchar *find)
+{
+    GList *header;
+    const gchar *const *pair;
+    
+    if (!(header = libbalsa_message_header_get_helper(headers, find)))
+        return NULL;
+
+    pair = header->data;
+    return pair[1];
+}
+
+GList*
+libbalsa_message_header_get_all(LibBalsaMessageHeaders* headers,
+                                const gchar *find)
+{
+    GList *header;
+    const gchar *const *pair;
+    GList *res = NULL;
+    
+    if (!(header = libbalsa_message_header_get_helper(headers, find)))
+        return NULL;
+    pair = header->data;
+    for(pair++; *pair; pair++)
+        res = g_list_append(res, g_strdup(*pair));
+    
+    return res;
+}
+
 const gchar *
 libbalsa_message_get_user_header(LibBalsaMessage * message,
                                  const gchar * name)
@@ -380,10 +420,11 @@ static void
 prepend_header_misc(const char *name, const char *value,
                     gpointer user_data)
 {
-    char lcname[17]; /* one byte longer than the longest ignored header */
+    char lcname[28]; /* one byte longer than the longest ignored header */
     static const char ignored_headers[] =
         "subject date from to cc bcc "
-        "message-id references in-reply-to status lines";
+        "message-id references in-reply-to status lines"
+        "disposition-notification-to";
     unsigned i;
     GList *res = *(GList **)user_data;
     if (!*value)
@@ -458,11 +499,18 @@ libbalsa_message_user_hdrs_from_gmime(GMimeMessage * message)
     value = g_mime_message_get_header(message, "In-Reply-To");
     if (value) {
         res =
+#if HAVE_GMIME_2_2_5
+            g_list_prepend(res,
+                           libbalsa_create_hdr_pair
+                           ("In-Reply-To",
+                            g_mime_utils_header_decode_text(value)));
+#else  /* HAVE_GMIME_2_2_5 */
             g_list_prepend(res,
                            libbalsa_create_hdr_pair
                            ("In-Reply-To",
                             g_mime_utils_header_decode_text
                             ((unsigned char *) value)));
+#endif /* HAVE_GMIME_2_2_5 */
     }
 
     g_mime_header_foreach(GMIME_OBJECT(message)->headers,
@@ -542,49 +590,22 @@ libbalsa_message_get_attach_icon(LibBalsaMessage * message)
 	return LIBBALSA_MESSAGE_ATTACH_ICONS_NUM;
 }
 
-/* Helper for mailbox drivers. */
-gboolean
-libbalsa_message_set_msg_flags(LibBalsaMessage * message,
-			       LibBalsaMessageFlag set,
-			       LibBalsaMessageFlag clear)
-{
-    gboolean changed = FALSE;
-    
-    if (!(message->flags & set)) {
-	message->flags |= set;
-	changed = TRUE;
-    }
-    if (message->flags & clear) {
-	message->flags &= ~clear;
-	changed = TRUE;
-    }
-
-    return changed;
-}
-
 /* Tell the mailbox driver to change flags. */
-static void
-libbalsa_message_set_flag(LibBalsaMessage * message,
-			  LibBalsaMessageFlag set,
-			  LibBalsaMessageFlag clear)
+void
+libbalsa_message_change_flags(LibBalsaMessage * message,
+                              LibBalsaMessageFlag set,
+                              LibBalsaMessageFlag clear)
 {
-    GArray *msgnos;
-
     if (message->mailbox->readonly) {
-	libbalsa_information(
-	    LIBBALSA_INFORMATION_WARNING,
-	    _("Mailbox (%s) is readonly: cannot change flags."),
-	    message->mailbox->name);
-	return;
+        libbalsa_information(LIBBALSA_INFORMATION_WARNING,
+                             _("Mailbox (%s) is readonly: "
+                               "cannot change flags."),
+                             message->mailbox->name);
+        return;
     }
 
-    msgnos = g_array_sized_new(FALSE, FALSE, sizeof(guint), 1);
-    g_array_append_val(msgnos, message->msgno);
-    libbalsa_mailbox_register_msgnos(message->mailbox, msgnos);
-    libbalsa_mailbox_messages_change_flags(message->mailbox, msgnos,
-					   set, clear);
-    libbalsa_mailbox_unregister_msgnos(message->mailbox, msgnos);
-    g_array_free(msgnos, TRUE);
+    libbalsa_mailbox_msgno_change_flags(message->mailbox, message->msgno,
+                                        set, clear);
 }
 
 void
@@ -592,59 +613,10 @@ libbalsa_message_reply(LibBalsaMessage * message)
 {
     g_return_if_fail(message->mailbox);
     libbalsa_lock_mailbox(message->mailbox);
-    libbalsa_message_set_flag(message, LIBBALSA_MESSAGE_FLAG_REPLIED, 0);
+    libbalsa_message_change_flags(message, LIBBALSA_MESSAGE_FLAG_REPLIED, 0);
     libbalsa_unlock_mailbox(message->mailbox);
 }
 
-/* Assume all messages come from the same mailbox */
-void
-libbalsa_messages_change_flag(GList * messages,
-                              LibBalsaMessageFlag flag,
-                              gboolean set)
-{
-    GArray *msgnos;
-    LibBalsaMessage * message = NULL;
-    LibBalsaMailbox *mbox;
-    
-    if (!messages)
-	return;
-
-    mbox = LIBBALSA_MESSAGE(messages->data)->mailbox;
-    if (mbox->readonly) {
-	libbalsa_information(
-	    LIBBALSA_INFORMATION_WARNING,
-	    _("Mailbox (%s) is readonly: cannot change flags."),
-	    mbox->name);
-	return;
-    }
-
-    /* Construct the list of messages that actually change state */
-    msgnos = g_array_new(FALSE, FALSE, sizeof(guint));
-    for (; messages; messages = messages->next) {
-	message = LIBBALSA_MESSAGE(messages->data);
- 	if ( (set && !(message->flags & flag)) ||
-             (!set && (message->flags & flag)) )
-	    g_array_append_val(msgnos, message->msgno);
-    }
-    libbalsa_mailbox_register_msgnos(message->mailbox, msgnos);
-    
-    if (msgnos->len > 0) {
-	libbalsa_lock_mailbox(mbox);
-	/* RETURN_IF_MAILBOX_CLOSED(mbox); */
-        /* set flags for entire set in one transaction */
-        if(set)
-            libbalsa_mailbox_messages_change_flags(mbox, msgnos,
-						   flag, 0);
-        else
-            libbalsa_mailbox_messages_change_flags(mbox, msgnos,
-						   0, flag);
-	libbalsa_unlock_mailbox(mbox);
-    }
-
-    libbalsa_mailbox_unregister_msgnos(message->mailbox, msgnos);
-    g_array_free(msgnos, TRUE);
-}
-
 
 /* libbalsa_message_body_ref:
    references the structure of given message possibly fetching also all
@@ -846,9 +818,7 @@ libbalsa_message_get_subject(LibBalsaMessage* msg)
        msg->mime_msg && msg->mailbox) { /* a message in a mailbox... */
         g_return_val_if_fail(MAILBOX_OPEN(msg->mailbox), NULL);
         ret = g_mime_message_get_subject(msg->mime_msg);
-	if (ret)
-	    LIBBALSA_MESSAGE_SET_SUBJECT(msg, (gchar*)
-				 ret = g_mime_utils_header_decode_text(ret));
+        libbalsa_message_set_subject_from_header(msg, ret);
     } else
 	ret = msg->subj;
 
@@ -993,12 +963,7 @@ libbalsa_message_init_from_gmime(LibBalsaMessage * message,
 
 #ifdef MESSAGE_COPY_CONTENT
     header = g_mime_message_get_subject(mime_msg);
-    if (header) {
-        message->subj =
-            g_mime_utils_header_decode_text((unsigned char *) header);
-        canonize_header_value(message->subj);
-	libbalsa_utf8_sanitize(&message->subj, TRUE, NULL);
-    }
+    libbalsa_message_set_subject_from_header(message, header);
 
     header = g_mime_message_get_header(mime_msg, "Content-Length");
     if (header)
@@ -1113,14 +1078,35 @@ static gboolean
 lbmsg_set_header(LibBalsaMessage *message, const gchar *name,
                  const gchar* value, gboolean all)
 {
+    gchar *val = NULL;
+
+    if (libbalsa_text_attr_string(value)) {
+        /* Broken header: force it to utf8 using Balsa's fallback
+         * charset, then rfc2047-encode it for passing to the
+         * appropriate GMime decoder. */
+        gchar *tmp = g_strdup(value);
+        libbalsa_utf8_sanitize(&tmp, TRUE, NULL);
+#if HAVE_GMIME_2_2_5
+        val = g_mime_utils_header_encode_text(tmp);
+#else  /* HAVE_GMIME_2_2_5 */
+        val = g_mime_utils_header_encode_text((unsigned char *) tmp);
+#endif /* HAVE_GMIME_2_2_5 */
+        g_free(tmp);
+#ifdef DEBUG
+        g_print("%s: non-ascii \"%s\" header \"%s\" encoded as \"%s\"\n",
+                __func__, name, value, val);
+#endif /* DEBUG */
+        value = val;
+    }
+
     if (g_ascii_strcasecmp(name, "Subject") == 0) {
-	if (!strcmp(value, "DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA"))
+	if (!strcmp(value,
+                    "DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA")) {
+            g_free(val);
 	    return FALSE;
+        }
 #if MESSAGE_COPY_CONTENT
-	g_free(message->subj);
-        message->subj =
-	    g_mime_utils_header_decode_text((unsigned char *) value);
-	libbalsa_utf8_sanitize(&message->subj, TRUE, NULL);
+        libbalsa_message_set_subject_from_header(message, value);
 #endif
     } else
     if (g_ascii_strcasecmp(name, "Date") == 0) {
@@ -1145,15 +1131,22 @@ lbmsg_set_header(LibBalsaMessage *message, const gchar *name,
 	message->headers->content_type =
 	    g_mime_content_type_new_from_string(value);
     } else
+    if (g_ascii_strcasecmp(name, "Disposition-Notification-To") == 0) {
+        message->headers->dispnotify_to =
+            internet_address_parse_string(value);
+    } else
 #ifdef MESSAGE_COPY_CONTENT
     if (g_ascii_strcasecmp(name, "Content-Length") == 0) {
 	    message->length = atoi(value);
     } else
 #endif
     if (all)
-	message->headers->user_hdrs =
-	    g_list_append(message->headers->user_hdrs,
-			  libbalsa_create_hdr_pair(name, g_strdup(value)));
+        message->headers->user_hdrs =
+            g_list_prepend(message->headers->user_hdrs,
+                           libbalsa_create_hdr_pair(name,
+                                                    g_strdup(value)));
+
+    g_free(val);
 
     return TRUE;
 }
@@ -1237,10 +1230,10 @@ libbalsa_message_load_envelope_from_stream(LibBalsaMessage * message,
 	line->len = 0;
 	g_byte_array_append(line, &lookahead, 1);
     } while (!g_mime_stream_eos(gmime_stream_buffer));
-    if (ret) {
-	/* calculate size */
-    }
+    if (ret)
+        message->length = g_mime_stream_length(gmime_stream);
     g_object_unref(gmime_stream_buffer);
+    g_mime_stream_reset(gmime_stream);
     libbalsa_mime_stream_shared_unlock(gmime_stream);
     g_byte_array_free(line, TRUE);
 }
@@ -1309,3 +1302,30 @@ libbalsa_message_copy(LibBalsaMessage * message, LibBalsaMailbox * dest,
 
     return retval;
 }
+
+void
+libbalsa_message_set_subject(LibBalsaMessage * message,
+                             const gchar * subject)
+{
+    g_free(message->subj);
+    message->subj = g_strdup(subject);
+    libbalsa_utf8_sanitize(&message->subj, TRUE, NULL);
+    canonize_header_value(message->subj);
+}
+
+void
+libbalsa_message_set_subject_from_header(LibBalsaMessage * message,
+                                         const gchar * header)
+{
+    if (header) {
+        gchar *subject =
+#if HAVE_GMIME_2_2_5
+            g_mime_utils_header_decode_text(header);
+#else  /* HAVE_GMIME_2_2_5 */
+            g_mime_utils_header_decode_text((const unsigned char *)
+                                            header);
+#endif /* HAVE_GMIME_2_2_5 */
+        libbalsa_message_set_subject(message, subject);
+        g_free(subject);
+    }
+}
