@@ -179,16 +179,16 @@ process_mime_multipart(LibBalsaMessage * message,
 }
 
 GString *
-content2reply(LibBalsaMessage * message, gchar * reply_prefix_str,
+content2reply(LibBalsaMessageBody * root, gchar * reply_prefix_str,
 	      gint llen, gboolean ignore_html, gboolean flow,
 	      LibBalsaCharsetFunc charset_cb, gpointer charset_cb_data)
 {
+    LibBalsaMessage *message = root->message;
     LibBalsaMessageBody *body;
     GString *reply = NULL, *res;
 
     libbalsa_message_body_ref(message, FALSE, FALSE);
-    body = message->body_list;
-    for (body = message->body_list; body; body = body->next) {
+    for (body = root; body; body = body->next) {
 	res = process_mime_part(message, body, reply_prefix_str, llen,
                                 ignore_html, flow, charset_cb, 
 				charset_cb_data);
@@ -269,7 +269,9 @@ unwrap_rfc2646(gchar * str, gboolean from_screen, gboolean delsp)
             for (p = str; *p == QUOTE_CHAR; p++)
                 /* nothing */;
             len = p - str;
-            sig_sep = (strncmp(p, "-- \n", 4) == 0);
+            sig_sep = (p[0] == '-' && p[1] == '-' && p[2] == ' '
+                       && (p[3] == '\n'
+                           || (p[3] == '\r' && p[4] == '\n')));
             if (text->quote_depth < 0)
                 text->quote_depth = len;
             else if (len != text->quote_depth || sig_sep)
@@ -1114,7 +1116,6 @@ libbalsa_unwrap_selection(GtkTextBuffer * buffer, regex_t * rex)
 	do
 	    gtk_text_buffer_insert(buffer, &start, ">", 1);
 	while (--quote_depth);
-	gtk_text_buffer_insert(buffer, &start, " ", 1);
     }
     g_free(line);
 
