@@ -192,20 +192,21 @@ send_message_info_destroy(SendMessageInfo *smi)
 
 
 #if HAVE_GPGME
-static gint
+static LibBalsaMsgCreateResult
 libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
-			       GtkWindow * parent);
+			       GtkWindow * parent, GError ** error);
 static LibBalsaMsgCreateResult
 do_multipart_crypto(LibBalsaMessage * message, GMimeObject ** mime_root,
-		    GtkWindow * parent);
+		    GtkWindow * parent, GError ** error);
 #endif
 
 static guint balsa_send_message_real(SendMessageInfo* info);
 static LibBalsaMsgCreateResult
 libbalsa_message_create_mime_message(LibBalsaMessage* message,
-				     gboolean flow, gboolean postponing);
+				     gboolean flow, gboolean postponing,
+				     GError ** error);
 static LibBalsaMsgCreateResult libbalsa_create_msg(LibBalsaMessage * message,
-				    gboolean flow);
+						   gboolean flow, GError ** error);
 static LibBalsaMsgCreateResult
 libbalsa_fill_msg_queue_item_from_queu(LibBalsaMessage * message,
                                        MessageQueueItem *mqi);
@@ -390,7 +391,7 @@ libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 #if ENABLE_ESMTP
                        LibBalsaSmtpServer * smtp_server,
 #endif /* ESMTP */
-		       gboolean flow)
+		       gboolean flow, GError ** error)
 {
     LibBalsaMsgCreateResult result;
 #if ENABLE_ESMTP
@@ -401,7 +402,7 @@ libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 
     g_return_val_if_fail(message, LIBBALSA_MESSAGE_CREATE_ERROR);
 
-    if ((result = libbalsa_create_msg(message, flow)) !=
+    if ((result = libbalsa_create_msg(message, flow, error)) !=
 	LIBBALSA_MESSAGE_CREATE_OK)
         return result;
 
@@ -458,18 +459,6 @@ libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
     return rc ?  LIBBALSA_MESSAGE_CREATE_OK : LIBBALSA_MESSAGE_QUEUE_ERROR;
 }
 
-static void
-lbs_change_flags(LibBalsaMessage * msg, LibBalsaMessageFlag set,
-                 LibBalsaMessageFlag clear)
-{
-    GArray *array = g_array_new(FALSE, FALSE, sizeof(guint));
-
-    g_array_append_val(array, msg->msgno);
-    libbalsa_mailbox_messages_change_flags(msg->mailbox, array, set,
-                                           clear);
-    g_array_free(array, TRUE);
-}
-
 /* libbalsa_message_send:
    send the given messsage (if any, it can be NULL) and all the messages
    in given outbox.
@@ -485,7 +474,8 @@ libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
                       LibBalsaMailbox * fccbox,
                       LibBalsaFccboxFinder finder,
                       LibBalsaSmtpServer * smtp_server,
-                      gboolean flow, gboolean debug)
+                      gboolean flow, gboolean debug,
+		      GError ** error)
 {
     LibBalsaMsgCreateResult result = LIBBALSA_MESSAGE_CREATE_OK;
 
@@ -494,7 +484,7 @@ libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 
     if (message != NULL)
         result = libbalsa_message_queue(message, outbox, fccbox,
-                                        smtp_server, flow);
+                                        smtp_server, flow, error);
 
     if (result == LIBBALSA_MESSAGE_CREATE_OK
         && !lbs_process_queue(outbox, finder, smtp_server, debug))
@@ -506,12 +496,13 @@ libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 LibBalsaMsgCreateResult
 libbalsa_message_send(LibBalsaMessage* message, LibBalsaMailbox* outbox,
 		      LibBalsaMailbox* fccbox, LibBalsaFccboxFinder finder,
-                      gboolean flow, gboolean debug)
+                      gboolean flow, gboolean debug,
+		      GError ** error)
 {
     LibBalsaMsgCreateResult result = LIBBALSA_MESSAGE_CREATE_OK;
 
     if (message != NULL)
- 	result = libbalsa_message_queue(message, outbox, fccbox, flow);
+ 	result = libbalsa_message_queue(message, outbox, fccbox, flow, error);
     if (result == LIBBALSA_MESSAGE_CREATE_OK)
  	if (!libbalsa_process_queue(outbox, finder, debug))
  	    return LIBBALSA_MESSAGE_SEND_ERROR;
@@ -629,14 +620,13 @@ lbs_process_queue(LibBalsaMailbox * outbox, LibBalsaFccboxFinder finder,
 	const gchar *smtp_server_name;
         LibBalsaMsgCreateResult created;
 
-	msg = libbalsa_mailbox_get_message(outbox, msgno);
-        if (LIBBALSA_MESSAGE_HAS_FLAG(msg,
-                                      (LIBBALSA_MESSAGE_FLAG_FLAGGED |
-                                       LIBBALSA_MESSAGE_FLAG_DELETED))) {
-	    g_object_unref(msg);
+        if (libbalsa_mailbox_msgno_has_flags(outbox, msgno, 
+                                             (LIBBALSA_MESSAGE_FLAG_FLAGGED |
+                                              LIBBALSA_MESSAGE_FLAG_DELETED),
+                                             0))
             continue;
-	}
 
+	msg = libbalsa_mailbox_get_message(outbox, msgno);
         libbalsa_message_body_ref(msg, TRUE, TRUE);
         smtp_server_name =
             libbalsa_message_get_user_header(msg, "X-Balsa-SmtpServer");
@@ -658,7 +648,8 @@ lbs_process_queue(LibBalsaMailbox * outbox, LibBalsaFccboxFinder finder,
 	} else {
 	    InternetAddress *ia;
 
-	    lbs_change_flags(msg, LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
+	    libbalsa_message_change_flags(msg,
+                                          LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	    /*
 	       The message needs to be filtered and the newlines converted to
 	       \r\n because internally the lines foolishly terminate with the
@@ -878,7 +869,7 @@ disp_recipient_status(smtp_recipient_t recipient,
 
   if(status->code != 0 && status->code != 250) {
       libbalsa_information(
-                           LIBBALSA_INFORMATION_WARNING, 
+                           LIBBALSA_INFORMATION_ERROR,
                            _("Could not send the message to %s:\n"
                              "%d: %s\n"
                              "Message left in your outbox.\n"), 
@@ -901,7 +892,8 @@ handle_successful_send(smtp_message_t message, void *be_verbose)
       mqi->refcount--;
 
     if(mqi != NULL && mqi->orig != NULL && mqi->orig->mailbox)
-	lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_FLAGGED);
+	libbalsa_message_change_flags(mqi->orig, 0,
+                                      LIBBALSA_MESSAGE_FLAG_FLAGGED);
     else printf("mqi: %p mqi->orig: %p mqi->orig->mailbox: %p\n",
                   mqi, mqi ? mqi->orig : NULL, 
                   mqi&&mqi->orig ? mqi->orig->mailbox : NULL);
@@ -916,8 +908,9 @@ handle_successful_send(smtp_message_t message, void *be_verbose)
 	    if (mqi->orig->mailbox && fccurl) {
                 LibBalsaMailbox *fccbox = mqi->finder(fccurl);
                 GError *err = NULL;
-                lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_NEW |
-                                 LIBBALSA_MESSAGE_FLAG_FLAGGED);
+                libbalsa_message_change_flags(mqi->orig, 0,
+                                              LIBBALSA_MESSAGE_FLAG_NEW |
+                                              LIBBALSA_MESSAGE_FLAG_FLAGGED);
 		libbalsa_mailbox_sync_storage(mqi->orig->mailbox, FALSE);
                 remove = libbalsa_message_copy(mqi->orig, fccbox, &err);
                 if(!remove) {
@@ -931,15 +924,14 @@ handle_successful_send(smtp_message_t message, void *be_verbose)
             /* If copy failed, mark the message again as flagged -
                otherwise it will get resent again. And again, and
                again... */
-            lbs_change_flags(mqi->orig, remove ?
-                             LIBBALSA_MESSAGE_FLAG_DELETED :
-                             LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
+            libbalsa_message_change_flags(mqi->orig, remove ?
+                                          LIBBALSA_MESSAGE_FLAG_DELETED :
+                                          LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	}
     } else {
-	/* XXX - Show the poor user the status codes and message. */
+	/* Check whether it was a problem with transfer. */
         if(*(gboolean*)be_verbose) {
             int cnt = 0;
-            /* status = smtp_reverse_path_status(message); */
             if(status->code != 250 && status->code != 0) {
                 libbalsa_information(
                                      LIBBALSA_INFORMATION_WARNING,
@@ -950,12 +942,21 @@ handle_successful_send(smtp_message_t message, void *be_verbose)
                 cnt++;
             }
             smtp_enumerate_recipients (message, disp_recipient_status, &cnt);
-            if(cnt==0) { /* other error */
-            libbalsa_information(
-           LIBBALSA_INFORMATION_WARNING,
-           _("Message submission problem, placing it into your outbox.\n"
-             "System will attempt to resubmit the message until you delete it."));
-
+            if(cnt==0) { /* other error, maybe sender or message size? */
+                status = smtp_reverse_path_status(message);
+                if(status->code != 250 && status->code != 0) {
+                    libbalsa_information
+                        (LIBBALSA_INFORMATION_WARNING,
+                         _("Relaying refused:\n"
+                           "%d: %s\n"
+                           "Message left in your outbox.\n"), 
+                         status->code, status->text);
+                } else
+                    libbalsa_information
+                        (LIBBALSA_INFORMATION_ERROR,
+                         _("Message submission problem, placing it into your outbox.\n"
+                           "System will attempt to resubmit the message until you delete it."));
+                
             }
         }
     }
@@ -999,7 +1000,7 @@ libbalsa_smtp_event_cb (smtp_session_t session, int event_no, void *arg, ...)
         snprintf (buf, sizeof buf, _("From %s: %d %s"), 
 	          mailbox, status->code, status->text);
         g_strchomp(buf);
-	libbalsa_information(LIBBALSA_INFORMATION_MESSAGE, buf);
+	libbalsa_information(LIBBALSA_INFORMATION_DEBUG, buf);
         break;
     case SMTP_EV_RCPTSTATUS:
         mailbox = va_arg (ap, const char *);
@@ -1014,7 +1015,7 @@ libbalsa_smtp_event_cb (smtp_session_t session, int event_no, void *arg, ...)
         snprintf (buf, sizeof buf, _("To %s: %d %s"),
 	          mailbox, status->code, status->text);
         g_strchomp(buf);
-	libbalsa_information(LIBBALSA_INFORMATION_MESSAGE, buf);
+	libbalsa_information(LIBBALSA_INFORMATION_DEBUG, buf);
         break;
     case SMTP_EV_MESSAGEDATA:
         message = va_arg (ap, smtp_message_t);
@@ -1039,7 +1040,7 @@ libbalsa_smtp_event_cb (smtp_session_t session, int event_no, void *arg, ...)
 	snprintf (buf, sizeof buf, "%d %s", status->code, status->text);
         g_strchomp(buf);
 	MSGSENDTHREAD(threadmsg, MSGSENDTHREADPROGRESS, buf, NULL, NULL, 0);
-	libbalsa_information(LIBBALSA_INFORMATION_MESSAGE, buf);
+	libbalsa_information(LIBBALSA_INFORMATION_DEBUG, buf);
         /* Reset 'mqi->sent' for the next message (i.e. bcc copy) */
         mqi = smtp_message_get_application_data (message);
         if (mqi != NULL) {
@@ -1173,13 +1174,14 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, LibBalsaFccboxFinder finder,
     for (msgno = libbalsa_mailbox_total_messages(outbox);
 	 msgno > 0; msgno--) {
         LibBalsaMsgCreateResult created;
-	msg = libbalsa_mailbox_get_message(outbox, msgno);
 
-        if (LIBBALSA_MESSAGE_HAS_FLAG(msg,
-                                      (LIBBALSA_MESSAGE_FLAG_FLAGGED |
-                                       LIBBALSA_MESSAGE_FLAG_DELETED)))
+        if (libbalsa_mailbox_msgno_has_flags(outbox, msgno, 
+                                             (LIBBALSA_MESSAGE_FLAG_FLAGGED |
+                                              LIBBALSA_MESSAGE_FLAG_DELETED),
+                                             0))
             continue;
 
+	msg = libbalsa_mailbox_get_message(outbox, msgno);
         libbalsa_message_body_ref(msg, TRUE, TRUE); /* FIXME: do we need
                                                       * all headers? */
 	new_message = msg_queue_item_new(finder);
@@ -1189,7 +1191,8 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, LibBalsaFccboxFinder finder,
 	if (created != LIBBALSA_MESSAGE_CREATE_OK) {
 	    msg_queue_item_destroy(new_message);
 	} else {
-	    lbs_change_flags(msg, LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
+	    libbalsa_message_change_flags(msg,
+                                          LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	    if (mqi)
 		mqi->next_message = new_message;
 	    else
@@ -1226,8 +1229,9 @@ handle_successful_send(MessageQueueItem *mqi, LibBalsaFccboxFinder finder)
         const gchar *fccurl =
             libbalsa_message_get_user_header(mqi->orig, "X-Balsa-Fcc");
 
-        lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_NEW |
-                         LIBBALSA_MESSAGE_FLAG_FLAGGED);
+        libbalsa_message_change_flags(mqi->orig, 0,
+                                      LIBBALSA_MESSAGE_FLAG_NEW |
+                                      LIBBALSA_MESSAGE_FLAG_FLAGGED);
 	libbalsa_mailbox_sync_storage(mqi->orig->mailbox, FALSE);
 
         if (mqi->orig->mailbox && fccurl) {
@@ -1238,9 +1242,9 @@ handle_successful_send(MessageQueueItem *mqi, LibBalsaFccboxFinder finder)
         /* If copy failed, mark the message again as flagged -
            otherwise it will get resent again. And again, and
            again... */
-        lbs_change_flags(mqi->orig, remove ?
-                         LIBBALSA_MESSAGE_FLAG_DELETED :
-                         LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
+        libbalsa_message_change_flags(mqi->orig, remove ?
+                                      LIBBALSA_MESSAGE_FLAG_DELETED :
+                                      LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
     }
     mqi->status = MQI_SENT;
 }
@@ -1566,7 +1570,7 @@ get_tz_offset(time_t *t)
 
 static LibBalsaMsgCreateResult
 libbalsa_message_create_mime_message(LibBalsaMessage* message, gboolean flow,
-				     gboolean postponing)
+				     gboolean postponing, GError ** error)
 {
     gchar **mime_type;
     GMimeObject *mime_root = NULL;
@@ -1696,16 +1700,16 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gboolean flow,
 	    if (!postponing && body == body->message->body_list &&
 		message->gpg_mode > 0 &&
 		(message->gpg_mode & LIBBALSA_PROTECT_OPENPGP) != 0) {
-		gint result = 
+		LibBalsaMsgCreateResult crypt_res =
 		    libbalsa_create_rfc2440_buffer(body,
 			                           GMIME_PART(mime_part),
-						   parent);
+						   parent, error);
 
-		if (result != LIBBALSA_MESSAGE_CREATE_OK) {
+		if (crypt_res != LIBBALSA_MESSAGE_CREATE_OK) {
 		    g_object_unref(G_OBJECT(mime_part));
 		    if (mime_root)
 			g_object_unref(G_OBJECT(mime_root));
-		    return LIBBALSA_MESSAGE_CREATE_ERROR;
+		    return crypt_res;
 		}
 	    }
 #else
@@ -1727,7 +1731,7 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gboolean flow,
 #ifdef HAVE_GPGME
     if (message->body_list != NULL && !postponing) {
 	LibBalsaMsgCreateResult crypt_res =
-	    do_multipart_crypto(message, &mime_root, parent);
+	    do_multipart_crypto(message, &mime_root, parent, error);
 	if (crypt_res != LIBBALSA_MESSAGE_CREATE_OK)
 	    return crypt_res;
     }
@@ -1826,19 +1830,26 @@ gboolean
 libbalsa_message_postpone(LibBalsaMessage * message,
                           LibBalsaMailbox * draftbox,
                           LibBalsaMessage * reply_message,
-                          gchar * fcc, gboolean flow)
+			  gchar ** extra_headers, gboolean flow)
 {
     gboolean retval;
     GError *err = NULL;
 
+    /* in postpone mode no crypto operation is triggered, so we don't need to
+       pass the error */
     if (!message->mime_msg
         && libbalsa_message_create_mime_message(message, flow,
-                                                TRUE) !=
+                                                TRUE, NULL) !=
         LIBBALSA_MESSAGE_CREATE_OK)
         return FALSE;
 
-    if (fcc)
-        g_mime_message_set_header(message->mime_msg, "X-Balsa-Fcc", fcc);
+    if (extra_headers) {
+	gint i;
+
+	for (i = 0; extra_headers[i] && extra_headers[i + 1]; i += 2)
+	    g_mime_message_set_header(message->mime_msg, extra_headers[i],
+				      extra_headers[i + 1]);
+    }
 
     retval = libbalsa_message_copy(message, draftbox, &err);
 
@@ -1890,12 +1901,12 @@ libbalsa_set_message_id(GMimeMessage * mime_message)
    copies message to msg.
 */ 
 static LibBalsaMsgCreateResult
-libbalsa_create_msg(LibBalsaMessage * message, gboolean flow)
+libbalsa_create_msg(LibBalsaMessage * message, gboolean flow, GError ** error)
 {
     if (!message->mime_msg) {
 	LibBalsaMsgCreateResult res =
 	    libbalsa_message_create_mime_message(message, flow,
-						 FALSE);
+						 FALSE, error);
 	if (res != LIBBALSA_MESSAGE_CREATE_OK)
 	    return res;
     }
@@ -1953,9 +1964,9 @@ lb_send_from(LibBalsaMessage *message)
     return ia->value.addr;
 }
 
-static gint
+static LibBalsaMsgCreateResult
 libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
-			       GtkWindow * parent)
+			       GtkWindow * parent, GError ** error)
 {
     LibBalsaMessage *message = body->message;
     gint mode = message->gpg_mode;
@@ -1967,7 +1978,7 @@ libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
 	    if (!libbalsa_rfc2440_sign_encrypt(mime_part, 
 					       lb_send_from(message),
 					       NULL, FALSE,
-					       parent))
+					       parent, error))
 		return LIBBALSA_MESSAGE_SIGN_ERROR;
 	    break;
 	case LIBBALSA_PROTECT_ENCRYPT:
@@ -1995,14 +2006,14 @@ libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
 					              lb_send_from(message),
 						      encrypt_for,
 						      always_trust,
-						      parent);
+						      parent, error);
 		else
 		    result = 
 			libbalsa_rfc2440_sign_encrypt(mime_part, 
 						      NULL,
 						      encrypt_for,
 						      always_trust,
-						      parent);
+						      parent, error);
 		g_list_foreach(encrypt_for, (GFunc) g_free, NULL);
 		g_list_free(encrypt_for);
 		if (!result)
@@ -2010,8 +2021,7 @@ libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
 	    }
 	    break;
 	default:
-	    g_error("illegal gpg_mode %d (" __FILE__ " line %d)",
-		    mode, __LINE__);
+	    g_assert_not_reached();
 	}
 
     return LIBBALSA_MESSAGE_CREATE_OK;
@@ -2021,7 +2031,7 @@ libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part,
 /* handle rfc2633 and rfc3156 signing and/or encryption of a message */
 static LibBalsaMsgCreateResult
 do_multipart_crypto(LibBalsaMessage * message, GMimeObject ** mime_root,
-		    GtkWindow * parent)
+		    GtkWindow * parent, GError ** error)
 {
     gpgme_protocol_t protocol;
     gboolean always_trust;
@@ -2049,7 +2059,7 @@ do_multipart_crypto(LibBalsaMessage * message, GMimeObject ** mime_root,
 	case LIBBALSA_PROTECT_SIGN:   /* sign message */
 	    if (!libbalsa_sign_mime_object(mime_root,
 					   lb_send_from(message),
-					   protocol, parent))
+					   protocol, parent, error))
 		return LIBBALSA_MESSAGE_SIGN_ERROR;
 	    break;
 	case LIBBALSA_PROTECT_ENCRYPT:
@@ -2076,12 +2086,13 @@ do_multipart_crypto(LibBalsaMessage * message, GMimeObject ** mime_root,
 			libbalsa_sign_encrypt_mime_object(mime_root,
 							  lb_send_from(message),
 							  encrypt_for, protocol,
-							  always_trust, parent);
+							  always_trust, parent,
+							  error);
 		else
 		    success = 
 			libbalsa_encrypt_mime_object(mime_root, encrypt_for,
 						     protocol, always_trust,
-						     parent);
+						     parent, error);
 		g_list_free(encrypt_for);
 		
 		if (!success)
