@@ -569,7 +569,7 @@ periodic_expunge_cb(void)
 #endif
 
     gdk_threads_enter();
-    libbalsa_information(LIBBALSA_INFORMATION_MESSAGE,
+    libbalsa_information(LIBBALSA_INFORMATION_DEBUG,
                          _("Compressing mail folders..."));
     gtk_tree_model_foreach(GTK_TREE_MODEL(balsa_app.mblist_tree_store),
 			   (GtkTreeModelForeachFunc)mbnode_expunge_func,
@@ -578,7 +578,8 @@ periodic_expunge_cb(void)
 
     for (l = list; l; l = l->next) {
         BalsaMailboxNode *mbnode = l->data;
-        if (mbnode->mailbox && libbalsa_mailbox_is_open(mbnode->mailbox)) {
+        if (mbnode->mailbox && libbalsa_mailbox_is_open(mbnode->mailbox)
+            && !mbnode->mailbox->readonly) {
             time_t tm = time(NULL);
             if (tm-mbnode->last_use > balsa_app.expunge_timeout)
                 libbalsa_mailbox_sync_storage(mbnode->mailbox, TRUE);
@@ -608,6 +609,12 @@ balsa_progress_set_text(LibBalsaProgress * progress, const gchar * text,
     gboolean rc = FALSE;
 
     gdk_threads_enter();
+
+    if (!balsa_app.main_window) {
+        gdk_threads_leave();
+        return;
+    }
+
     if (!text || total >= LIBBALSA_PROGRESS_MIN_COUNT)
         rc = balsa_window_setup_progress(balsa_app.main_window, text);
     prev_fraction = 0;
@@ -631,8 +638,9 @@ balsa_progress_set_fraction(LibBalsaProgress * progress, gdouble fraction)
     prev_fraction = fraction;
 
     gdk_threads_enter();
-    balsa_window_increment_progress(balsa_app.main_window, fraction,
-                                    !libbalsa_am_i_subthread());
+    if (balsa_app.main_window)
+        balsa_window_increment_progress(balsa_app.main_window, fraction,
+                                        !libbalsa_am_i_subthread());
     gdk_threads_leave();
 }
 
@@ -750,7 +758,7 @@ main(int argc, char *argv[])
         BalsaSendmsg *snd;
         GSList *lst;
         gdk_threads_enter();
-        snd = sendmsg_window_new(window, NULL, SEND_NORMAL);
+        snd = sendmsg_window_compose(window);
         gdk_threads_leave();
         if(opt_compose_email) {
             if(g_ascii_strncasecmp(opt_compose_email, "mailto:", 7) == 0)
