@@ -352,7 +352,7 @@ prepare_header_real(PrintInfo * pi, LibBalsaMessageBody * sig_body,
     }
     g_free(subject);
 
-    date = libbalsa_message_headers_date_to_gchar(headers, balsa_app.date_string);
+    date = libbalsa_message_headers_date_to_utf8(headers, balsa_app.date_string);
     print_header_string (&pdata->headers, "date", _("Date:"), date);
     g_free(date);
 
@@ -473,7 +473,7 @@ prepare_message_header(PrintInfo * pi, LibBalsaMessageBody * body)
 	footer_string = g_string_new(subject);
     g_free(subject);
 
-    date = libbalsa_message_date_to_gchar(pi->message, balsa_app.date_string);
+    date = libbalsa_message_date_to_utf8(pi->message, balsa_app.date_string);
     if (footer_string) {
 	footer_string = g_string_append(footer_string, " - ");
 	footer_string = g_string_append(footer_string, date);
@@ -645,6 +645,7 @@ prepare_html(PrintInfo * pi, LibBalsaMessageBody * body, gpointer data)
     gchar *html_text;
     gchar *conttype;
     LibBalsaHTMLType html_type;
+    GError *err = NULL;
 
     conttype = libbalsa_message_body_get_mime_type(body);
     html_type = libbalsa_html_type(conttype);
@@ -668,9 +669,14 @@ prepare_html(PrintInfo * pi, LibBalsaMessageBody * body, gpointer data)
 	return;
     }
 
-    len = libbalsa_message_body_get_content(body, &html_text);
-    if (!html_text)
+    len = libbalsa_message_body_get_content(body, &html_text, &err);
+    if (!html_text) {
+        balsa_information(LIBBALSA_INFORMATION_ERROR,
+                          _("Could not get a part: %s"),
+                          err ? err->message : "Unknown error");
+        g_clear_error(&err);
 	return;
+    }
 
     len = libbalsa_html_filter(html_type, &html_text, len);
 
@@ -867,7 +873,7 @@ prepare_plaintext(PrintInfo * pi, LibBalsaMessageBody * body, gpointer data)
     if (body->buffer)
 	textbuf = g_strdup(body->buffer);
     else
-	libbalsa_message_body_get_content(body, &textbuf);
+	libbalsa_message_body_get_content(body, &textbuf, NULL);
 
     /* fake an empty buffer if textbuf is NULL */
     if (!textbuf)
