@@ -536,7 +536,8 @@ config_mailbox_init(const gchar * prefix)
 	if (special) {
 	    /* Designate as special before appending to the mblist, so
 	     * that view->show gets set correctly for sentbox, draftbox,
-	     * and outbox. */
+	     * and outbox, and view->subscribe gets set correctly for
+             * trashbox. */
 	    *special = mailbox;
 	    g_object_add_weak_pointer(G_OBJECT(mailbox), (gpointer) special);
 	}
@@ -586,12 +587,81 @@ config_load_smtp_server(const gchar * key, const gchar * value, gpointer data)
     libbalsa_conf_push_group(key);
     smtp_server = libbalsa_smtp_server_new_from_config(value);
     libbalsa_conf_pop_group();
-    *smtp_servers = g_slist_prepend(*smtp_servers, smtp_server);
+    libbalsa_smtp_server_add_to_list(smtp_server, smtp_servers);
 
     return FALSE;
 }
 #endif                          /* ENABLE_ESMTP */
 
+#ifdef HAVE_GTK_PRINT
+static gboolean
+load_gtk_print_setting(const gchar * key, const gchar * value, gpointer data)
+{
+    g_return_val_if_fail(data != NULL, TRUE);
+    gtk_print_settings_set((GtkPrintSettings *)data, key, value);
+    return FALSE;
+}
+
+static GtkPageSetup *
+restore_gtk_page_setup()
+{
+    GtkPageSetup *page_setup;
+    GtkPaperSize *paper_size = NULL;
+    gdouble width;
+    gdouble height;
+    gchar *name;
+    gchar *ppd_name;
+    gchar *display_name;
+
+    width = libbalsa_conf_get_double("Width");
+    height = libbalsa_conf_get_double("Height");
+
+    name = libbalsa_conf_get_string("PaperName");
+    ppd_name = libbalsa_conf_get_string ("PPDName");
+    display_name = libbalsa_conf_get_string("DisplayName");
+    page_setup = gtk_page_setup_new();
+
+    if (ppd_name && *ppd_name)
+	paper_size = gtk_paper_size_new_from_ppd(ppd_name, display_name,
+						 width, height);
+    else if (name && *name)
+	paper_size = gtk_paper_size_new_custom(name, display_name,
+					       width, height, GTK_UNIT_MM);
+
+    /* set defaults if no info is available */
+    if (!paper_size) {
+	paper_size = gtk_paper_size_new(GTK_PAPER_NAME_A4);
+	gtk_page_setup_set_paper_size_and_default_margins(page_setup,
+							  paper_size);
+    } else {
+	gtk_page_setup_set_paper_size(page_setup, paper_size);
+
+	gtk_page_setup_set_top_margin(page_setup,
+				      libbalsa_conf_get_double("MarginTop"),
+				      GTK_UNIT_MM);
+	gtk_page_setup_set_bottom_margin(page_setup,
+					 libbalsa_conf_get_double("MarginBottom"),
+					 GTK_UNIT_MM);
+	gtk_page_setup_set_left_margin(page_setup,
+				       libbalsa_conf_get_double("MarginLeft"),
+				       GTK_UNIT_MM);
+	gtk_page_setup_set_right_margin(page_setup,
+					libbalsa_conf_get_double("MarginRight"),
+					GTK_UNIT_MM);
+	gtk_page_setup_set_orientation(page_setup,
+				       (GtkPageOrientation)
+				       libbalsa_conf_get_int("Orientation"));
+    }
+    
+    gtk_paper_size_free(paper_size);
+    g_free(ppd_name);
+    g_free(name);
+    g_free(display_name);
+    
+    return page_setup;
+}
+#endif
+
 static gint
 config_global_load(void)
 {
@@ -830,6 +900,15 @@ config_global_load(void)
     libbalsa_conf_push_group("Printing");
 
     /* ... Printing */
+#ifdef HAVE_GTK_PRINT
+    if (balsa_app.page_setup)
+	g_object_unref(G_OBJECT(balsa_app.page_setup));
+    balsa_app.page_setup = restore_gtk_page_setup();
+    balsa_app.margin_left = libbalsa_conf_get_double("LeftMargin");
+    balsa_app.margin_top = libbalsa_conf_get_double("TopMargin");
+    balsa_app.margin_right = libbalsa_conf_get_double("RightMargin");
+    balsa_app.margin_bottom = libbalsa_conf_get_double("BottomMargin");
+#else
     g_free(balsa_app.paper_size);
     balsa_app.paper_size =
 	libbalsa_conf_get_string("PaperSize=" DEFAULT_PAPER_SIZE);
@@ -849,6 +928,7 @@ config_global_load(void)
     balsa_app.paper_orientation  = libbalsa_conf_get_string("PaperOrientation");
     g_free(balsa_app.page_orientation);
     balsa_app.page_orientation   = libbalsa_conf_get_string("PageOrientation");
+#endif
 
     g_free(balsa_app.print_header_font);
     balsa_app.print_header_font =
@@ -864,13 +944,26 @@ config_global_load(void)
                                 DEFAULT_PRINT_BODY_FONT);
     balsa_app.print_highlight_cited =
         libbalsa_conf_get_bool("PrintHighlightCited=false");
+#ifdef HAVE_GTK_PRINT
+    balsa_app.print_highlight_phrases =
+        libbalsa_conf_get_bool("PrintHighlightPhrases=false");
+#endif
     libbalsa_conf_pop_group();
 
+#ifdef HAVE_GTK_PRINT
+    /* GtkPrint printing */
+    libbalsa_conf_push_group("GtkPrint");
+    libbalsa_conf_foreach_keys("GtkPrint", load_gtk_print_setting,
+			       balsa_app.print_settings);
+    libbalsa_conf_pop_group();
+#endif
+
     /* Spelling options ... */
     libbalsa_conf_push_group("Spelling");
 
 #if HAVE_GTKSPELL
-    balsa_app.spell_check = libbalsa_conf_get_bool("SpellCheck=false");
+    balsa_app.spell_check_lang =
+        libbalsa_conf_get_string("SpellCheckLanguage");
 #else                           /* HAVE_GTKSPELL */
     balsa_app.module = d_get_gint("PspellModule", DEFAULT_PSPELL_MODULE);
     balsa_app.suggestion_mode =
@@ -986,7 +1079,7 @@ config_global_load(void)
                                        &def_used);
     if (!def_used)
         g_idle_add((GSourceFunc) config_warning_idle,
-                   _("The option not to send \"format=flowed\" is now "
+                   _("The option not to send \"format=flowed\" text is now "
                      "on the Options menu of the compose window."));
 
     balsa_app.autoquote = 
@@ -1095,6 +1188,54 @@ config_global_load(void)
     return TRUE;
 }				/* config_global_load */
 
+#ifdef HAVE_GTK_PRINT
+static void
+save_gtk_print_setting(const gchar *key, const gchar *value, gpointer user_data)
+{
+    libbalsa_conf_set_string(key, value);
+}
+
+static void
+save_gtk_page_setup(GtkPageSetup *page_setup)
+{
+    GtkPaperSize *paper_size;
+
+    if (!page_setup)
+	return;
+
+    paper_size = gtk_page_setup_get_paper_size(page_setup);
+    g_assert(paper_size != NULL);
+
+    libbalsa_conf_set_string("PaperName",
+			     gtk_paper_size_get_name (paper_size));
+    libbalsa_conf_set_string("DisplayName",
+			     gtk_paper_size_get_display_name (paper_size));
+    libbalsa_conf_set_string("PPDName",
+			     gtk_paper_size_get_ppd_name (paper_size));
+
+    libbalsa_conf_set_double("Width",
+			     gtk_paper_size_get_width(paper_size,
+						      GTK_UNIT_MM));
+    libbalsa_conf_set_double("Height",
+			     gtk_paper_size_get_height(paper_size,
+						       GTK_UNIT_MM));
+    libbalsa_conf_set_double("MarginTop",
+			     gtk_page_setup_get_top_margin(page_setup,
+							   GTK_UNIT_MM));
+    libbalsa_conf_set_double("MarginBottom",
+			     gtk_page_setup_get_bottom_margin(page_setup,
+							      GTK_UNIT_MM));
+    libbalsa_conf_set_double("MarginLeft",
+			     gtk_page_setup_get_left_margin(page_setup,
+							    GTK_UNIT_MM));
+    libbalsa_conf_set_double("MarginRight",
+			     gtk_page_setup_get_right_margin(page_setup,
+							     GTK_UNIT_MM));
+    libbalsa_conf_set_int("Orientation",
+			  gtk_page_setup_get_orientation(page_setup));
+}
+#endif
+
 gint
 config_save(void)
 {
@@ -1220,6 +1361,13 @@ config_save(void)
 
     /* Printing options ... */
     libbalsa_conf_push_group("Printing");
+#ifdef HAVE_GTK_PRINT
+    save_gtk_page_setup(balsa_app.page_setup);
+    libbalsa_conf_set_double("LeftMargin", balsa_app.margin_left);
+    libbalsa_conf_set_double("TopMargin", balsa_app.margin_top);
+    libbalsa_conf_set_double("RightMargin", balsa_app.margin_right);
+    libbalsa_conf_set_double("BottomMargin", balsa_app.margin_bottom);
+#else
     libbalsa_conf_set_string("PaperSize",balsa_app.paper_size);
     if(balsa_app.margin_left)
 	libbalsa_conf_set_string("LeftMargin",   balsa_app.margin_left);
@@ -1238,6 +1386,7 @@ config_save(void)
 				balsa_app.paper_orientation);
     if(balsa_app.margin_bottom)
 	libbalsa_conf_set_string("PageOrientation", balsa_app.page_orientation);
+#endif
 
     libbalsa_conf_set_string("PrintHeaderFont",
                             balsa_app.print_header_font);
@@ -1246,14 +1395,30 @@ config_save(void)
                             balsa_app.print_footer_font);
     libbalsa_conf_set_bool("PrintHighlightCited",
                           balsa_app.print_highlight_cited);
+#ifdef HAVE_GTK_PRINT
+    libbalsa_conf_set_bool("PrintHighlightPhrases",
+                          balsa_app.print_highlight_phrases);
+#endif
     libbalsa_conf_pop_group();
 
+#ifdef HAVE_GTK_PRINT
+    /* GtkPrintSettings stuff */
+    libbalsa_conf_remove_group("GtkPrint");
+    libbalsa_conf_push_group("GtkPrint");
+    if (balsa_app.print_settings)
+	gtk_print_settings_foreach(balsa_app.print_settings,
+				   save_gtk_print_setting, NULL);
+    libbalsa_conf_pop_group();
+#endif
+
     /* Spelling options ... */
+    libbalsa_conf_remove_group("Spelling");
     libbalsa_conf_push_group("Spelling");
 
 #if HAVE_GTKSPELL
-    libbalsa_conf_set_bool("SpellCheck", balsa_app.spell_check);
-    balsa_app.spell_check = libbalsa_conf_get_bool("SpellCheck=false");
+    if (balsa_app.spell_check_lang)
+        libbalsa_conf_set_string("SpellCheckLanguage",
+                                 balsa_app.spell_check_lang);
 #else                           /* HAVE_GTKSPELL */
     libbalsa_conf_set_int("PspellModule", balsa_app.module);
     libbalsa_conf_set_int("PspellSuggestMode", balsa_app.suggestion_mode);
@@ -1661,6 +1826,10 @@ config_view_load(const gchar * key, const gchar * value, gpointer data)
         if (!def)
             view->show = tmp;
 
+        tmp = libbalsa_conf_get_int_with_default("Subscribe", &def);
+        if (!def)
+            view->subscribe = tmp;
+
         tmp = libbalsa_conf_get_bool_with_default("Exposed", &def);
         if (!def)
             view->exposed = tmp;
@@ -1746,10 +1915,10 @@ save_view(const gchar * url, LibBalsaMailboxView * view)
 	libbalsa_conf_set_int("SortType",    view->sort_type);
     if (view->sort_field     != libbalsa_mailbox_get_sort_field(NULL))
 	libbalsa_conf_set_int("SortField",   view->sort_field);
-    /* initial value for show is UNSET, but that's always replaced, 
-     * and we want it to default to FROM. */
-    if (view->show           != LB_MAILBOX_SHOW_FROM)
+    if (view->show           == LB_MAILBOX_SHOW_TO)
 	libbalsa_conf_set_int("Show",        view->show);
+    if (view->subscribe      == LB_MAILBOX_SUBSCRIBE_NO)
+	libbalsa_conf_set_int("Subscribe",   view->subscribe);
     if (view->exposed        != libbalsa_mailbox_get_exposed(NULL))
 	libbalsa_conf_set_bool("Exposed",    view->exposed);
     if (view->open           != libbalsa_mailbox_get_open(NULL))
