@@ -61,6 +61,8 @@ static void store_address_add_address(StoreAddressInfo * info,
                                       const gchar * label,
                                       const InternetAddress * address,
                                       const InternetAddress * group);
+static void store_address_add_lbaddress(StoreAddressInfo * info,
+                                        const LibBalsaAddress *address);
 static void store_address_add_list(StoreAddressInfo * info,
                                    const gchar * label,
 				   const InternetAddressList * list);
@@ -70,7 +72,7 @@ static void store_address_add_list(StoreAddressInfo * info,
  */
 #define BALSA_STORE_ADDRESS_KEY "balsa-store-address"
 void
-balsa_store_address(GList * messages)
+balsa_store_address_from_messages(GList * messages)
 {
     StoreAddressInfo *info = NULL;
     GList *message_list = NULL;
@@ -118,6 +120,24 @@ balsa_store_address(GList * messages)
     gtk_widget_show(GTK_WIDGET(info->dialog));
 }
 
+void
+balsa_store_address(const LibBalsaAddress *address)
+{
+    StoreAddressInfo *info = NULL;
+
+    info = g_new(StoreAddressInfo, 1);
+    info->current_address_book = NULL;
+    info->message_list = NULL;
+    info->entries_list = NULL;
+    info->dialog = store_address_dialog(info);
+
+    store_address_add_lbaddress(info, address);
+
+    g_signal_connect(G_OBJECT(info->dialog), "response",
+                     G_CALLBACK(store_address_response), info);
+    gtk_widget_show_all(GTK_WIDGET(info->dialog));
+}
+
 /* Weak notify that a message was deleted; remove it from our list. */
 static void
 store_address_weak_notify(StoreAddressInfo * info, gpointer message)
@@ -254,7 +274,7 @@ store_address_book_frame(StoreAddressInfo * info)
     GtkWidget *frame = gtk_frame_new(_("Choose Address Book"));
     GtkWidget *combo_box;
     LibBalsaAddressBook *address_book;
-    guint default_ab_offset = 0;
+    guint default_ab_offset = 0, off;
 
     combo_box = gtk_combo_box_new_text();
     g_signal_connect(combo_box, "changed",
@@ -262,10 +282,12 @@ store_address_book_frame(StoreAddressInfo * info)
     if (balsa_app.address_book_list) {
 	info->current_address_book = balsa_app.default_address_book;
 
-	
-	for(ab_list = balsa_app.address_book_list;
+	/* NOTE: we have to store the default address book index and
+           call set_active() after all books are added to the list or
+           gtk-2.10.4 will lose the setting. */
+	for(off=0, ab_list = balsa_app.address_book_list;
             ab_list; 
-            ab_list = g_list_next(ab_list)) {
+            off++, ab_list = g_list_next(ab_list)) {
 	    address_book = LIBBALSA_ADDRESS_BOOK(ab_list->data);
 	    if (info->current_address_book == NULL)
 		info->current_address_book = address_book;
@@ -273,12 +295,10 @@ store_address_book_frame(StoreAddressInfo * info)
 	    gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box),
                                       address_book->name);
 	    if (address_book == balsa_app.default_address_book)
-		gtk_combo_box_set_active(GTK_COMBO_BOX(combo_box),
-                                         default_ab_offset);
-
-	    default_ab_offset++;
-
+                default_ab_offset = off;
 	}
+        gtk_combo_box_set_active(GTK_COMBO_BOX(combo_box),
+                                 default_ab_offset);
     }
     gtk_container_add(GTK_CONTAINER(frame), combo_box);
 
@@ -360,6 +380,29 @@ store_address_add_address(StoreAddressInfo * info,
     g_free(label_text);
 }
 
+static void
+store_address_add_lbaddress(StoreAddressInfo * info,
+                            const LibBalsaAddress *address)
+{
+    gchar *label_text;
+    GtkWidget **entries, *ew;
+
+    g_return_if_fail(address->address_list);
+    entries = g_new(GtkWidget *, NUM_FIELDS);
+    info->entries_list = g_list_append(info->entries_list, entries);
+
+    ew = libbalsa_address_get_edit_widget(address, entries, NULL, NULL);
+
+    label_text = g_strdup(address->full_name ? address->full_name :
+                          address->address_list->data);
+    if (g_utf8_strlen(label_text, -1) > 10)
+        /* truncate to an arbitrary length: */
+        *g_utf8_offset_to_pointer(label_text, 10) = '\0';
+    gtk_notebook_append_page(GTK_NOTEBOOK(info->notebook), ew,
+                             gtk_label_new(label_text));
+    g_free(label_text);
+}
+
 /* store_address_add_list:
  * take a list of addresses and pass them one at a time to
  * store_address_add_address */
